"""
Callback functions
"""
import os
from typing import Dict

import tensorflow as tf


class ManualStop(tf.keras.callbacks.Callback):
    """
    Stop the training manually by putting a "STOP" file in the directory
    """

    def on_batch_end(self, epoch: int, logs: Dict = None) -> None:
        """
        Codes called at the end of a batch
        Args:
            epoch (int): epoch id
            logs (Dict): log dict

        Returns: None

        """
        if os.path.isfile("STOP"):
            self.model.stop_training = True



# -*- coding: utf-8 -*-
import unittest

import numpy as np

from m3gnet.graph import Index
from m3gnet.layers import GraphUpdateFunc


class TestAgg(unittest.TestCase):
    def test_agg(self):
        graph = [[] for _ in range(13)]
        graph[Index.ATOMS] = np.random.normal(size=(10, 3))
        graph[Index.N_ATOMS] = np.array([5, 5])
        graph[Index.STATES] = np.random.normal(size=(2, 5))
        guf = GraphUpdateFunc(update_func=lambda x: x * 100, update_field="atoms")
        graph2 = guf(graph)
        self.assertTrue(np.allclose(graph2[Index.ATOMS], graph[Index.ATOMS] * 100))


if __name__ == "__main__":
    unittest.main()



"""Define several types for convenient use"""

from typing import Any, Callable, List, Optional, Union

import numpy as np
from ase.atoms import Atoms
from pymatgen.core import Molecule, Structure

OptStrOrCallable = Optional[Union[str, Callable[..., Any]]]
StructureOrMolecule = Union[Structure, Molecule, Atoms]
VectorLike = Union[List[float], np.ndarray]



import unittest

import numpy as np
import tensorflow as tf
from pymatgen.core import Lattice, Structure

from m3gnet.graph import (
    Index,
    RadiusCutoffGraphConverter,
    get_pair_vector_from_graph,
    tf_compute_distance_angle,
)


def _loop_indices(bond_atom_indices, pair_dist, cutoff=4.0):
    bin_count = np.bincount(bond_atom_indices[:, 0], minlength=bond_atom_indices[-1, 0] + 1)
    indices = []
    start = 0
    for bcont in bin_count:
        for i in range(bcont):
            for j in range(bcont):
                if start + i == start + j:
                    continue
                if pair_dist[start + i] > cutoff or pair_dist[start + j] > cutoff:
                    continue
                indices.append([start + i, start + j])
        start += bcont
    return np.array(indices)


def _calculate_cos_loop(graph, threebody_cutoff=4.0):
    """
    Calculate the cosine theta of triplets using loops
    Args:
        graph: List
    Returns: a list of cosine theta values
    """
    pair_vector = get_pair_vector_from_graph(graph)
    _, _, n_sites = tf.unique_with_counts(graph[Index.BOND_ATOM_INDICES][:, 0])
    start_index = 0
    cos = []
    for n_site in n_sites:
        for i in range(n_site):
            for j in range(n_site):
                if i == j:
                    continue
                vi = pair_vector[i + start_index].numpy()
                vj = pair_vector[j + start_index].numpy()
                di = np.linalg.norm(vi)
                dj = np.linalg.norm(vj)
                if (di <= threebody_cutoff) and (dj <= threebody_cutoff):
                    cos.append(vi.dot(vj) / np.linalg.norm(vi) / np.linalg.norm(vj))
        start_index += n_site
    return cos


class TestCompute(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        gc = RadiusCutoffGraphConverter(5)
        cls.s1 = Structure(Lattice.cubic(3.17), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        cls.s2 = Structure(Lattice.cubic(3), ["Mo", "Fe"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        cls.g1 = gc.convert(cls.s1)
        cls.g2 = gc.convert(cls.s2)

    def test_compute_pair_vec(self):
        pair_vec = get_pair_vector_from_graph(self.g1.as_list())

        d = tf.linalg.norm(pair_vec, axis=1)
        _, _, _, d2 = self.s1.get_neighbor_list(r=5.0)

        np.testing.assert_array_almost_equal(np.sort(d), np.sort(d2))

    def test_compute_dist_angle(self):
        g2 = tf_compute_distance_angle(
            self.g1.as_list(),
        )
        _, _, _, d2 = self.s1.get_neighbor_list(r=5.0)

        np.testing.assert_array_almost_equal(np.sort(g2[Index.BONDS].numpy().ravel()), np.sort(d2))

        cos_loop = _calculate_cos_loop(self.g1.as_list())

        cos = g2[Index.THETA]
        np.testing.assert_array_almost_equal(cos_loop, cos)

    def test_include_threebody_indices(self):
        g3 = self.g1.as_list()[:]
        np.testing.assert_array_almost_equal(
            g3[Index.TRIPLE_BOND_INDICES],
            _loop_indices(g3[Index.BOND_ATOM_INDICES], g3[Index.BONDS].ravel()),
        )


if __name__ == "__main__":
    unittest.main()



"""Data types"""
import numpy as np
import tensorflow as tf

DTYPES = {
    "float32": {"numpy": np.float32, "tf": tf.float32},
    "float16": {"numpy": np.float16, "tf": tf.float16},
    "int32": {"numpy": np.int32, "tf": tf.int32},
    "int16": {"numpy": np.int16, "tf": tf.int16},
}


class DataType:
    """
    Tensorflow and numpy data types. Used to choose between float16 and float32
    """

    np_float = tf.keras.mixed_precision.global_policy().compute_dtype
    np_int = "int32"
    tf_float = tf.keras.mixed_precision.global_policy().compute_dtype
    tf_int = "int32"

    @classmethod
    def set_dtype(cls, data_type: str) -> None:
        """
        Class method to set the data types
        Args:
            data_type (str): '16' or '32'
        """
        if data_type.endswith("32"):
            float_key = "float32"
            int_key = "int32"
        elif data_type.endswith("16"):
            float_key = "float16"
            int_key = "int16"
        else:
            raise ValueError("Data type not known, choose '16' or '32'")

        cls.np_float = DTYPES[float_key]["numpy"]
        cls.tf_float = DTYPES[float_key]["tf"]
        cls.np_int = DTYPES[int_key]["numpy"]
        cls.tf_int = DTYPES[int_key]["tf"]


def set_global_dtypes(data_type) -> None:
    """
    Function to set the data types
    Args:
        data_type (str): '16' or '32'
    Returns:

    """
    DataType.set_dtype(data_type)



"""
The core m3gnet model
"""
import json
import logging
import os
import urllib.request
from typing import List, Optional

import numpy as np
import tensorflow as tf

from m3gnet.graph import Index, RadiusCutoffGraphConverter, tf_compute_distance_angle
from m3gnet.layers import (
    MLP,
    AtomReduceState,
    AtomRef,
    BaseAtomRef,
    ConcatAtoms,
    ConcatBondAtomState,
    GatedAtomUpdate,
    GatedMLP,
    GraphFeaturizer,
    GraphNetworkLayer,
    GraphUpdateFunc,
    MultiFieldReadout,
    Pipe,
    ReduceReadOut,
    Set2Set,
    SphericalBesselWithHarmonics,
    StateNetwork,
    ThreeDInteraction,
    WeightedReadout,
    polynomial,
)
from m3gnet.utils import register_plain

from ._base import GraphModelMixin

logger = logging.getLogger(__file__)

CWD = os.path.dirname(os.path.abspath(__file__))

"""
# Pre-trained models naming guidelines

To ensure clarity on the training data on the models, the naming convention for pre-trained models should be
<source>-<date in YYYY.MM.DD format>-E(F)(S), where the E, F and S denotes energies, forces and stresses respectively.
For example, MP-2021.2.8-EFS denotes a potential trained on Materials Project energies, forces and stresses as of
2021.2.8.
"""
MODEL_NAME = "m3gnet"

MODEL_PATHS = {"MP-2021.2.8-EFS": os.path.join(CWD, "..", "..", "pretrained", "MP-2021.2.8-EFS")}

MODEL_FILES = {
    "MP-2021.2.8-EFS": ["checkpoint", "m3gnet.json", "m3gnet.index", "m3gnet.data-00000-of-00001"],
}

GITHUB_RAW_LINK = "https://raw.githubusercontent.com/materialsvirtuallab/m3gnet/main/pretrained/{model_name}/{filename}"


def _download_file(url: str, target: str):
    logger.info(f"Downloading {target} from {url} ... ")
    if not os.path.isfile(target):
        urllib.request.urlretrieve(url, target)


def _download_model_to_dir(model_name: str = "MP-2021.2.8-EFS", dirname: str = "MP-2021.2.8-EFS"):
    if model_name not in MODEL_FILES:
        raise ValueError(f"{model_name} not supported. Currently we only have {MODEL_FILES.keys()}")
    full_dirname = os.path.join(CWD, dirname)
    if not os.path.isdir(full_dirname):
        os.mkdir(full_dirname)
    for filename in MODEL_FILES[model_name]:
        _download_file(
            GITHUB_RAW_LINK.format(model_name=model_name, filename=filename), os.path.join(full_dirname, filename)
        )
    logger.info(f"Model {model_name} downloaded to {full_dirname}")


@register_plain
class M3GNet(GraphModelMixin, tf.keras.models.Model):
    """
    The main M3GNet model
    """

    def __init__(
        self,
        max_n: int = 3,
        max_l: int = 3,
        n_blocks: int = 3,
        units: int = 64,
        cutoff: float = 5.0,
        threebody_cutoff: float = 4.0,
        n_atom_types: int = 94,
        include_states: bool = False,
        readout: str = "weighted_atom",
        task_type: str = "regression",
        is_intensive: bool = True,
        mean: float = 0.0,
        std: float = 1.0,
        element_refs: Optional[np.ndarray] = None,
        **kwargs,
    ):
        r"""
        Args:
            max_n (int): number of radial basis expansion
            max_l (int): number of angular expansion
            n_blocks (int): number of convolution blocks
            units (int): number of neurons in each MLP layer
            cutoff (float): cutoff radius of the graph
            threebody_cutoff (float): cutoff radius for 3 body interaction
            n_atom_types (int): number of atom types
            include_states (bool): whether to include states calculation
            readout (str): the readout function type. choose from `set2set`,
                `weighted_atom` and `reduce_atom`, default to `weighted_atom`
            task_type (str): `classification` or `regression`, default to
                `regression`
            is_intensive (bool): whether the prediction is intensive
            mean (float): optional `mean` value of the target
            std (float): optional `std` of the target
            element_refs (np.ndarray): element reference values for each
                element
            **kwargs:
        """
        super().__init__(**kwargs)
        self.graph_converter = RadiusCutoffGraphConverter(cutoff=cutoff, threebody_cutoff=threebody_cutoff)

        if include_states:
            self.graph_converter.set_default_states(np.array([[0.0, 0.0]], dtype="float32"))

        if task_type.lower() == "classification":
            act_final = "sigmoid"
        else:
            act_final = None

        self.featurizer = GraphFeaturizer(
            n_atom_types=n_atom_types,
            atom_embedding_dim=units,
            rbf_type="SphericalBessel",
            max_n=max_n,
            max_l=max_l,
            cutoff=cutoff,
            smooth=True,
        )

        self.feature_adjust = GraphUpdateFunc(MLP([units], activations=["swish"], use_bias=False), "bonds")

        self.basis_expansion = SphericalBesselWithHarmonics(max_n=max_n, max_l=max_l, cutoff=cutoff, use_phi=False)
        update_size = max_n * max_l

        self.three_interactions = [
            ThreeDInteraction(
                update_network=MLP([update_size], activations=["sigmoid"]),
                update_network2=GatedMLP([units], activations=["swish"], use_bias=False),
            )
            for _ in range(n_blocks)
        ]

        self.graph_layers = []

        for i in range(n_blocks):
            atom_network = GatedAtomUpdate(neurons=[units, units], activation="swish")

            bond_network = ConcatAtoms(neurons=[units, units], activation="swish")

            if include_states:
                atom_agg_func = AtomReduceState()
                state_network = ConcatBondAtomState(
                    update_func=MLP([units, units], activations=["swish", "swish"]),
                    atom_agg_func=atom_agg_func,
                    bond_agg_func=None,
                )
            else:
                state_network = StateNetwork()

            layer = GraphNetworkLayer(
                atom_network=atom_network,
                bond_network=bond_network,
                state_network=state_network,
            )
            self.graph_layers.append(layer)

        if is_intensive:
            if readout == "set2set":
                atom_readout = Set2Set(units=units, num_steps=2, field="atoms")

            elif readout == "weighted_atom":
                atom_readout = WeightedReadout(neurons=[units, units], field="atoms")
            else:
                atom_readout = ReduceReadOut("mean", field="atoms")

            readout_nn = MultiFieldReadout(atom_readout=atom_readout, include_states=include_states)

            mlp = MLP([units, units, 1], ["swish", "swish", act_final], is_output=True)

            self.final = Pipe(layers=[readout_nn, mlp])

        else:
            if task_type == "classification":
                raise ValueError("Classification task cannot be extensive")
            final_layers = []
            if include_states:
                final_layers.append(
                    GraphNetworkLayer(atom_network=GatedAtomUpdate(neurons=[units], activation="swish"))
                )

            final_layers.append(
                GraphNetworkLayer(
                    atom_network=GraphUpdateFunc(
                        update_func=GatedMLP(
                            neurons=[units, units, 1],
                            activations=["swish", "swish", None],
                        ),
                        update_field="atoms",
                    )
                )
            )
            final_layers.append(ReduceReadOut(method="sum", field="atoms"))
            self.final = Pipe(layers=final_layers)

        if element_refs is None:
            self.element_ref_calc = BaseAtomRef()
        else:
            self.element_ref_calc = AtomRef(property_per_element=element_refs, max_z=n_atom_types)
        self.max_n = max_n
        self.max_l = max_l
        self.n_blocks = n_blocks
        self.n_atom_types = n_atom_types
        self.units = units
        self.cutoff = cutoff
        self.threebody_cutoff = threebody_cutoff
        self.include_states = include_states
        self.readout = readout
        self.task_type = task_type
        self.is_intensive = is_intensive
        self.kwargs = kwargs
        self.mean = mean
        self.std = std
        self.element_refs = element_refs

    def call(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns:

        """
        graph = tf_compute_distance_angle(graph)
        property_offset = self.element_ref_calc(graph)
        three_basis = self.basis_expansion(graph)
        three_cutoff = polynomial(graph[Index.BONDS], self.threebody_cutoff)
        g = self.featurizer(graph)
        g = self.feature_adjust(g)
        for i in range(self.n_blocks):
            g = self.three_interactions[i](g, three_basis, three_cutoff)
            g = self.graph_layers[i](g)
        g = self.final(g)
        g = g * self.std + self.mean
        g += property_offset
        return g

    def get_config(self):
        """
        Get config dict for serialization
        Returns:
        """
        config = {"name": self.name}
        config.update(
            {
                "max_n": self.max_n,
                "max_l": self.max_l,
                "n_blocks": self.n_blocks,
                "units": self.units,
                "cutoff": self.cutoff,
                "threebody_cutoff": self.threebody_cutoff,
                "include_states": self.include_states,
                "readout": self.readout,
                "n_atom_types": self.n_atom_types,
                "task_type": self.task_type,
                "is_intensive": self.is_intensive,
                "mean": self.mean,
                "std": self.std,
                "element_refs": self.element_refs,
            }
        )
        return config

    @classmethod
    def from_config(cls, config: dict) -> "M3GNet":
        r"""
        Construct the model from a config dict
        Args:
            config (dict): config dict from `get_config` method
        Returns: new M3GNet instance
        """
        return cls(**config)

    def save(self, dirname: str):
        """
        Saves the model to a directory.

        Args:
            dirname (str): directory to save the model
        """
        model_serialized = self.to_json()
        model_name = os.path.join(dirname, MODEL_NAME)
        self.save_weights(model_name)
        if not os.path.isdir(dirname):
            os.mkdir(dirname)
        fname = os.path.join(dirname, MODEL_NAME + ".json")
        with open(fname, "w") as f:
            json.dump(model_serialized, f)

    @classmethod
    def from_dir(cls, dirname: str, custom_objects: Optional[dict] = None) -> "M3GNet":
        """
        Load the model from a directory

        Args:
            dirname (str): directory to save the model
            custom_objects (dict): dictionary for custom object
        Returns: M3GNet model
        """
        custom_objects = custom_objects or {}
        model_name = os.path.join(dirname, MODEL_NAME)
        fname = os.path.join(dirname, MODEL_NAME + ".json")
        if not os.path.isfile(fname):
            raise ValueError("Model does not exists")
        with open(fname, "r") as f:
            model_serialized = json.load(f)
        # model_serialized = _replace_compatibility(model_serialized)
        model = tf.keras.models.model_from_json(model_serialized, custom_objects=custom_objects)
        model.load_weights(model_name)
        return model

    def set_element_refs(self, element_refs: np.ndarray):
        """
        Set element reference for the property
        Args:
            element_refs (np.ndarray): element reference value for the
                extensive property
        """
        self.element_refs = element_refs
        self.element_ref_calc = AtomRef(property_per_element=element_refs)

    @classmethod
    def load(cls, model_name: str = "MP-2021.2.8-EFS") -> "M3GNet":
        """
        Load the model weights from pre-trained model
        Args:
            model_name (str): model name or the path for saved model. Defaults to "MP-2021.2.8-EFS".

        Returns: M3GNet object.
        """
        if model_name in MODEL_PATHS:
            try:
                return cls.from_dir(MODEL_PATHS[model_name])
            except ValueError:
                _download_model_to_dir(model_name, model_name)
                return cls.load(os.path.join(CWD, model_name))

        if os.path.isdir(model_name):
            if "m3gnet.json" in os.listdir(model_name):
                return cls.from_dir(model_name)

        raise ValueError(f"{model_name} not found in available pretrained {list(MODEL_PATHS.keys())}")



# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: profile=True
# cython: language_level=3
# distutils: language = c
# distutils: define_macros=NPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION

cimport numpy as np
import numpy as np
from libc.stdlib cimport free, malloc
from libc.string cimport memset


def compute_threebody(const int[:, ::1] bond_atom_indices,
    const int[::1] n_atoms):
    """
    Calculate the three body indices from pair atom indices

    Args:
        bond_atom_indices (np.ndarray): pair atom indices
        n_atoms (int): number of atoms

    Returns:
        triple_bond_indices (np.ndarray): bond indices that form three-body
        py_n_triple_ij (np.ndarray): number of three-body angles for each bond
        py_n_triple_i (np.ndarray): number of three-body angles each atom
        py_n_triple_s (np.ndarray): number of three-body angles for each
            structure
    """
    cdef int i, j, k
    cdef int n_bond = bond_atom_indices.shape[0]
    cdef int n_atom = 0
    cdef int n_struct = n_atoms.shape[0]
    for i in range(n_struct):
        n_atom += n_atoms[i]

    cdef int* n_bond_per_atom = <int *> malloc(n_atom * sizeof(int))
    memset(n_bond_per_atom, 0, n_atom * sizeof(int))

    for i in range(n_bond):
        n_bond_per_atom[bond_atom_indices[i, 0]] += 1

    cdef int* n_triple_i = <int *> malloc(n_atom * sizeof(int))
    cdef int* n_triple_ij = <int *> malloc(n_bond * sizeof(int))
    cdef int* n_triple_s = <int *> malloc(n_struct * sizeof(int))

    memset(n_triple_s, 0, n_struct * sizeof(int))

    cdef int n_triple = 0
    cdef int n_triple_temp
    cdef int start = 0

    for i in range(n_atom):
        n_triple_temp = n_bond_per_atom[i] * (n_bond_per_atom[i] - 1)
        for j in range(n_bond_per_atom[i]):
            n_triple_ij[start + j] = n_bond_per_atom[i] - 1
        n_triple += n_triple_temp
        n_triple_i[i] = n_triple_temp
        start += n_bond_per_atom[i]

    cdef np.ndarray triple_bond_indices = np.empty(shape=(n_triple, 2),
                                             dtype=np.int32)

    start = 0
    cdef int index = 0
    for i in range(n_atom):
        for j in range(n_bond_per_atom[i]):
            for k in range(n_bond_per_atom[i]):
                if j != k:
                    triple_bond_indices[index, 0] = start + j
                    triple_bond_indices[index, 1] = start + k
                    index += 1
        start += n_bond_per_atom[i]

    start = 0
    cdef int end = start
    cdef int n_atom_temp
    for i in range(n_struct):
        end += n_atoms[i]
        for j in range(start, end):
            n_triple_s[i] += n_triple_i[j]
        start = end
    py_n_triple_ij = np.array(<int[:n_bond]>n_triple_ij)
    py_n_triple_i = np.array(<int[:n_atom]>n_triple_i)
    py_n_triple_s = np.array(<int[:n_struct]>n_triple_s)

    free(n_triple_ij)
    free(n_triple_i)
    free(n_triple_s)
    free(n_bond_per_atom)
    return triple_bond_indices, py_n_triple_ij, py_n_triple_i, py_n_triple_s



"""
Base layer classes
"""
from typing import Callable, List, Optional

import tensorflow as tf

from m3gnet.graph import Index
from m3gnet.utils import register


@register
class GraphUpdate(tf.keras.layers.Layer):
    """
    A graph update takes a graph list representation as input and output a
    updated graph
    """

    def call(self, graph: List, **kwargs) -> List:
        """

        Args:
            graph (list): list representation of a graph
            **kwargs:
        Returns:
        """
        return graph


@register
class GraphUpdateFunc(GraphUpdate):
    """
    Update a graph with a function
    """

    def __init__(
        self,
        update_func: Callable,
        update_field: str,
        input_fields: Optional[List[str]] = None,
        **kwargs,
    ):
        """

        Args:
            update_func (Callable): A update function that is callable
            update_field (str): The graph field to update
            input_fields (list): by default only the "update_field" is the
             input to the update function. Alternatively we can specify
             multiple fields as inputs
            **kwargs:
        """
        super().__init__(**kwargs)
        self.update_func = update_func
        self.update_field: str = update_field
        if input_fields is None:
            input_fields = [update_field]  # type: ignore
        self.input_fields: List[str] = input_fields

    def call(self, graph: List, **kwargs) -> List:
        """
        Call logic of the layer
        Args:
            graph (list): list representation of the graph
            **kwargs:
        Returns: graph list
        """
        graph = graph[:]
        index = getattr(Index, self.update_field.upper())
        input_indices = [getattr(Index, i.upper()) for i in self.input_fields]
        graph[index] = self.update_func(*[graph[i] for i in input_indices])
        return graph

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: config dict
        """
        config = super().get_config()
        config.update(
            {
                "update_func": self.update_func,
                "update_field": self.update_field,
                "input_fields": self.input_fields,
            }
        )
        return config



import unittest

import numpy as np
from pymatgen.core import Lattice, Structure

from m3gnet.graph import BaseGraphConverter, MaterialGraph, RadiusCutoffGraphConverter


class TestConverter(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s1 = Structure(Lattice.cubic(3.17), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])

    def test_base_converter(self):
        bc = BaseGraphConverter()
        g1 = bc.get_atom_features(self.s1)
        np.testing.assert_array_equal(g1, [42, 42])
        bc.set_default_states(states=[0, 0])
        state = bc.get_states(self.s1)
        self.assertListEqual(state, [0, 0])
        self.s1.states = np.array([[0, 1]])
        state = bc.get_states(self.s1)
        np.testing.assert_array_almost_equal(state.ravel(), [0, 1])

    def test_radius_cutoff(self):
        rcg = RadiusCutoffGraphConverter(cutoff=4.0, has_threebody=True, threebody_cutoff=3.0)
        g1 = rcg.convert(self.s1)
        self.assertTrue(isinstance(g1, MaterialGraph))
        self.assertTrue(g1.has_threebody)
        self.assertTrue(np.all(g1.bonds.ravel() <= 4.0))

        rcg2 = RadiusCutoffGraphConverter(cutoff=4.0, has_threebody=False, threebody_cutoff=3.0)
        g2 = rcg2.convert(self.s1)
        self.assertFalse(g2.has_threebody)


if __name__ == "__main__":
    unittest.main()



"""
Three body basis expansion
"""
from typing import List

import tensorflow as tf

from m3gnet.graph import Index
from m3gnet.utils import (
    SphericalBesselFunction,
    SphericalHarmonicsFunction,
    combine_sbf_shf,
    register,
)


@register
class SphericalBesselWithHarmonics(tf.keras.layers.Layer):
    """
    Spherical bessel function as radial function and spherical harmonics as
    the angular function
    """

    def __init__(
        self, max_n: int, max_l: int, cutoff: float = 5.0, use_phi: bool = False, smooth: bool = False, **kwargs
    ):
        """

        Args:
            max_n (int): maximum number for radial expansion
            max_l (int): maximum number of angular expansion
            cutoff (float): cutoff radius
            use_phi (bool): whether to use phi. So far, we do not have the phi
                angle yet.
            smooth (bool): whether to use the smooth version of the radial
                function
            **kwargs:
        """
        super().__init__(**kwargs)
        self.sbf = SphericalBesselFunction(max_l=max_l, max_n=max_n, cutoff=cutoff, smooth=smooth)
        self.shf = SphericalHarmonicsFunction(max_l=max_l, use_phi=use_phi)
        self.max_n = max_n
        self.max_l = max_l
        self.cutoff = cutoff
        self.use_phi = use_phi

    def call(self, graph: List, **kwargs) -> tf.Tensor:  # noqa
        """
        Args:
            graph (list): the list representation of a graph
            **kwargs:
        Returns: combined radial and spherical harmonic expansion of the
            distance and angle
        """
        sbf = self.sbf(graph[Index.TRIPLE_BOND_LENGTHS])
        shf = self.shf(graph[Index.THETA], graph[Index.PHI])
        return combine_sbf_shf(sbf, shf, max_n=self.max_n, max_l=self.max_l, use_phi=self.use_phi)

    def get_config(self) -> dict:
        """
        Get the config dict for serialization
        Returns: config dict
        """
        config = super().get_config()
        config.update(
            {
                "max_l": self.max_l,
                "max_n": self.max_n,
                "cutoff": self.cutoff,
                "use_phi": self.use_phi,
            }
        )
        return config



# -*- coding: utf-8 -*-
import unittest

import numpy as np

from m3gnet.layers import RadialBasisFunctions


class TestRBF(unittest.TestCase):
    def test_rbf(self):
        r = np.linspace(1, 5, 11)
        rbf_gaussian = RadialBasisFunctions(rbf_type="Gaussian", centers=np.linspace(0, 5, 10), width=0.5)
        rbf = rbf_gaussian(r)
        self.assertTupleEqual(tuple(rbf.shape), (11, 10))
        rbf_sb = RadialBasisFunctions(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=5.0, smooth=False)
        rbf = rbf_sb(r)
        self.assertTupleEqual(tuple(rbf.shape), (11, 9))

        rbf_sb = RadialBasisFunctions(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=5.0, smooth=True)
        rbf = rbf_sb(r)
        self.assertTupleEqual(tuple(rbf.shape), (11, 3))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

import numpy as np
from pymatgen.core import Lattice, Molecule, Structure

from m3gnet.graph import MaterialGraph, RadiusCutoffGraphConverter
from m3gnet.layers import MLP
from m3gnet.layers._atom import GatedAtomUpdate
from m3gnet.layers._bond import ConcatAtoms
from m3gnet.layers._gn import GraphFeaturizer, GraphNetworkLayer
from m3gnet.layers._state import ConcatBondAtomState


class TestGN(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])
        s.states = np.array([[0.1, 0.2, 0.3, 0.4, 0.5]])
        mol = Molecule(["C", "O"], [[0, 0, 0], [1.1, 0, 0]])
        rc = RadiusCutoffGraphConverter(cutoff=4.0, has_threebody=True, threebody_cutoff=4.0)
        cls.struct_graph = rc.convert(s).as_list()
        cls.mol_graph = rc.convert(mol).as_list()

    def test_gn(self):
        bond_network = ConcatAtoms(neurons=[10, 9])
        atom_network = GatedAtomUpdate(neurons=[10, 5])
        state_network = ConcatBondAtomState(update_func=MLP(neurons=[5, 3]))
        gn = GraphNetworkLayer(
            bond_network=bond_network,
            atom_network=atom_network,
            state_network=state_network,
        )

        gf = GraphFeaturizer(
            n_atom_types=94,
            atom_embedding_dim=5,
            nfeat_state=5,
            rbf_type="SphericalBessel",
            max_n=3,
            max_l=3,
            cutoff=4.0,
            smooth=False,
        )

        g1 = gf(self.struct_graph)
        g2 = MaterialGraph.from_list(g1)
        self.assertTupleEqual(tuple(g2.atoms.shape), (2, 5))
        self.assertTupleEqual(tuple(g2.bonds.shape), (28, 9))
        g3 = MaterialGraph.from_list(gn(g2.as_list()))
        self.assertTupleEqual(tuple(g3.atoms.shape), (2, 5))
        self.assertTupleEqual(tuple(g3.bonds.shape), (28, 9))
        self.assertTupleEqual(tuple(g3.states.shape), (1, 3))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import os
import unittest

import numpy as np
from monty.tempfile import ScratchDir
from pymatgen.core import Lattice, Molecule, Structure

from m3gnet.models import M3GNet, MolecularDynamics, Potential, Relaxer


class TestModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.model = M3GNet.load()
        cls.potential = Potential(model=cls.model)
        cls.mol = Molecule(["C", "O"], [[0, 0, 0], [1.5, 0, 0]])
        cls.structure = Structure(Lattice.cubic(3.30), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])

    def test_m3gnet(self):

        g = self.model.graph_converter(self.mol)

        val = self.model.predict_structure(self.mol).numpy().ravel()
        val_graph = self.model.predict_graph(g).numpy().ravel()

        self.assertTrue(val.size == 1)
        self.assertAlmostEqual(val, val_graph)

        vals = self.model.predict_structures([self.mol, self.mol]).numpy().ravel()
        vals_graph = self.model.predict_graphs([g, g]).numpy().ravel()

        self.assertTrue(np.allclose(vals, [val, val]))
        self.assertTrue(np.allclose(vals_graph, [val, val]))

    def test_potential(self):
        e, f, s = self.potential.get_efs(self.structure)
        self.assertAlmostEqual(e.numpy().item(), -21.3307, 3)
        self.assertTrue(np.allclose(f.numpy().ravel(), np.zeros(shape=(2, 3)).ravel(), atol=1e-3))
        self.assertTrue(
            np.allclose(
                np.diag(s[0].numpy()).ravel(),
                np.array([[28.519585, 28.519585, 28.519585]]),
                atol=1e-2,
            )
        )

    def test_relaxer(self):
        relaxer = Relaxer()  # this loads the default model

        relax_results = relaxer.relax(self.structure)

        final_structure = relax_results["final_structure"]
        final_energy = relax_results["trajectory"].energies[-1] / 2
        self.assertAlmostEqual(final_structure.lattice.abc[0], 3.169, 2)
        self.assertAlmostEqual(final_energy.item(), -10.859, 3)

    def test_md(self):
        with ScratchDir("."):
            md = MolecularDynamics(atoms=self.structure, temperature=300, logfile="mo.log")

            md.run(10)
            self.assertTrue(os.path.isfile("mo.log"))


if __name__ == "__main__":
    unittest.main()



"""basis functions"""

import tensorflow as tf

from m3gnet.utils import Gaussian, SphericalBesselFunction

RBF_ALLOWED = {
    "Gaussian": {"params": ["centers", "width"], "class": Gaussian},
    "SphericalBessel": {
        "params": ["max_l", "max_n", "cutoff", "smooth"],
        "class": SphericalBesselFunction,
    },
}


class RadialBasisFunctions(tf.keras.layers.Layer):
    """
    Radial distribution function basis
    """

    def __init__(self, rbf_type: str = "Gaussian", **kwargs):
        """
        Args:
            rbf_type (str): RBF function type, choose between `Gaussian` or
                `SphericalBessel`
            **kwargs:
        """
        name = kwargs.pop("name", None)
        super().__init__(name=name)
        allowed_kwargs = {}
        keys = []
        for i, j in kwargs.items():
            if i in RBF_ALLOWED[rbf_type]["params"]:  # type: ignore
                allowed_kwargs.update({i: j})
                keys.append(i)

        missing_keys: set[str] = set(RBF_ALLOWED[rbf_type]["params"]) - set(keys)  # type: ignore
        if len(missing_keys) > 0:
            raise ValueError("kwargs ", missing_keys, " not present")
        self.allowed_kwargs = allowed_kwargs
        self.func = RBF_ALLOWED[rbf_type]["class"](**self.allowed_kwargs)  # type: ignore
        self.rbf_type = rbf_type

    def call(self, r: tf.Tensor, **kwargs):
        """
        Args:
            r (tf.Tensor): 1D radial distance tensor
            **kwargs:
        Returns: radial basis functions

        """
        return self.func(r)

    def get_config(self) -> dict:
        """
        Get config of the class for serialization
        Returns: dict

        """
        config = super().get_config()
        config.update({"rbf_type": self.rbf_type})
        config.update(self.allowed_kwargs)
        return config



# -*- coding: utf-8 -*-
"""
Utilities for m3gnet
"""

from ._general import check_array_equal, check_shape_consistency, reshape_array
from ._math import (
    Gaussian,
    SphericalBesselFunction,
    SphericalHarmonicsFunction,
    combine_sbf_shf,
    get_spherical_bessel_roots,
    spherical_bessel_roots,
    spherical_bessel_smooth,
)
from ._tf import (
    broadcast_states_to_atoms,
    broadcast_states_to_bonds,
    get_length,
    get_range_indices_from_n,
    get_segment_indices_from_n,
    register,
    register_plain,
    repeat_with_n,
    unsorted_segment_fraction,
    unsorted_segment_softmax,
)

__all__ = [
    "get_length",
    "get_segment_indices_from_n",
    "get_range_indices_from_n",
    "repeat_with_n",
    "unsorted_segment_softmax",
    "unsorted_segment_fraction",
    "broadcast_states_to_atoms",
    "broadcast_states_to_bonds",
    "register",
    "register_plain",
    "check_array_equal",
    "check_shape_consistency",
    "reshape_array",
    "get_spherical_bessel_roots",
    "spherical_bessel_roots",
    "Gaussian",
    "SphericalBesselFunction",
    "SphericalHarmonicsFunction",
    "combine_sbf_shf",
    "spherical_bessel_smooth",
]



# -*- coding: utf-8 -*-
"""
Tensorflow related utility
"""
from typing import List

import tensorflow as tf


def get_segment_indices_from_n(ns: tf.Tensor) -> tf.Tensor:
    """
    Get segment indices from number array. For example if
    ns = [2, 3], then the function will return [0, 0, 1, 1, 1]

    Args:
        ns: tf.Tensor, the number of atoms/bonds array

    Returns:
        object:

    Returns: segment indices tensor
    """
    return tf.repeat(tf.range(tf.shape(ns)[0]), repeats=ns)


def get_range_indices_from_n(ns: tf.Tensor) -> tf.Tensor:
    """
    Give ns = [2, 3], return [0, 1, 0, 1, 2]
    Args:
        ns: tf.Tensor, the number of atoms/bonds array

    Returns: range indices
    """
    max_n = tf.reduce_max(ns)
    n = tf.shape(ns)[0]
    n_range = tf.range(max_n)
    matrix = tf.tile(n_range[None, ...], [n, 1])
    mask = tf.sequence_mask(ns, max_n)
    return tf.cast(tf.boolean_mask(matrix, mask), tf.int32)


def repeat_with_n(tensor: tf.Tensor, n: tf.Tensor) -> tf.Tensor:
    """
    Repeat the first dimension according to n array.
    The
    Args:
        tensor: Tensor, tensor to augment
        n: Tensor, array to specify the repeat times for
            each item in the first dimension of tensor

    Returns: repeated tensor

    """
    return tf.repeat(tensor, repeats=n, axis=0)


def get_length(t):
    """
    equivalent to tf.linalg.norm(t, axis=1), but faster
    Args:
        t: (tf.Tensor), two dimensional tensor

    Returns: tf.Tensor
    """
    return tf.sqrt(tf.reduce_sum(t * t, axis=1))


def unsorted_segment_softmax(data, segment_ids, num_segments, weights=None):
    """
    Unsorted segment softmax with optional weights
    Args:
        data (tf.Tensor): original data
        segment_ids (tf.Tensor): tensor segment ids
        num_segments (int): number of segments
    Returns: tf.Tensor
    """
    data = tf.cast(data, "float32")
    if weights is None:
        weights = tf.constant(1.0, dtype="float32")
    else:
        weights = tf.cast(weights, dtype="float32")

    segment_max = tf.math.unsorted_segment_max(data, segment_ids, num_segments)
    maxes = tf.gather(segment_max, segment_ids)
    data -= maxes
    exp = tf.exp(data) * tf.squeeze(weights)

    softmax = tf.math.divide_no_nan(
        exp,
        tf.gather(tf.math.unsorted_segment_sum(exp, segment_ids, num_segments), segment_ids),
    )
    return tf.cast(softmax, tf.keras.mixed_precision.global_policy().compute_dtype)


def unsorted_segment_fraction(data, segment_ids, num_segments):
    """
    Segment softmax
    Args:
        data (tf.Tensor): original data
        segment_ids (tf.Tensor): segment ids
        num_segments (tf.Tensor): number of segments
    Returns:
    """
    data = tf.cast(data, "float32")
    segment_sum = tf.math.unsorted_segment_sum(data, segment_ids, num_segments)
    sums = tf.gather(segment_sum, segment_ids)
    data = tf.math.divide_no_nan(data, sums)
    return tf.cast(data, tf.keras.mixed_precision.global_policy().compute_dtype)


def broadcast_states_to_bonds(graph: List):
    """
    Broadcast state attributes of shape [Ns, Nstate] to
    bond attributes shape [Nb, Nstate]
    Args:
        graph: List

    Returns: broadcasted state attributes

    """
    from m3gnet.graph import Index

    return tf.repeat(graph[Index.STATES], graph[Index.N_BONDS], axis=0)


def broadcast_states_to_atoms(graph: List):
    """
    Broadcast state attributes of shape [Ns, Nstate] to
    bond attributes shape [Na, Nstate]
    Args:
        graph: List

    Returns: broadcasted state attributes

    """
    from m3gnet.graph import Index

    return tf.repeat(graph[Index.STATES], graph[Index.N_ATOMS], axis=0)


def register(cls):
    """
    Register m3gnet classes
    Args:
        cls: class name
    Returns:
    """
    return tf.keras.utils.register_keras_serializable(package="m3gnet")(cls)


def register_plain(cls):
    """register a simple class_name: instance pair to the custom object

    Args:
        cls: class name
    """
    tf.keras.utils.get_custom_objects()[cls.__name__] = cls
    return cls



# -*- coding: utf-8 -*-
"""
math functions and miscellaneous calculations
"""

import os
from functools import lru_cache
from math import pi
from typing import List, Optional, Union

import numpy as np
import sympy
import tensorflow as tf
from scipy.optimize import brentq
from scipy.special import spherical_jn

from m3gnet.config import DataType

CWD = os.path.dirname(os.path.abspath(__file__))


@lru_cache(maxsize=128)
def get_spherical_bessel_roots() -> np.ndarray:
    """
    Get precomputed Spherical Bessel function roots. The results is a
    2D matrix with dimension [128, 128]. The n-th (0-based index） root of
    order l Spherical Bessel function is the `[l, n]` entry

    Returns: 2D matrix of pre-computed roots

    """
    return np.loadtxt(os.path.join(CWD, "sb_roots.txt"))


@lru_cache(maxsize=128)
def spherical_bessel_roots(max_l: int, max_n: int):
    """
    Calculate the spherical Bessel roots. The n-th root of the l-th
    spherical bessel function is the `[l, n]` entry of the return matrix.
    The calculation is based on the fact that the n-root for l-th
    spherical Bessel function `j_l`, i.e., `z_{j, n}` is in the range
    `[z_{j-1,n}, z_{j-1, n+1}]`. On the other hand we know precisely the
    roots for j0, i.e., sinc(x)

    Args:
        max_l: max order of spherical bessel function
        max_n: max number of roots
    Returns: root matrix of size [max_l, max_n]
    """
    temp_zeros = np.arange(1, max_l + max_n + 1) * pi  # j0
    roots = [temp_zeros[:max_n]]
    for i in range(1, max_l):
        roots_temp = []
        for j in range(max_n + max_l - i):
            low = temp_zeros[j]
            high = temp_zeros[j + 1]
            root = brentq(lambda x, v: spherical_jn(v, x), low, high, (i,))
            roots_temp.append(root)
        temp_zeros = np.array(roots_temp)
        roots.append(temp_zeros[:max_n])
    return np.array(roots)


class SphericalBesselFunction:
    """
    Calculate the spherical Bessel function based on the sympy + tensorflow
    implementations
    """

    def __init__(self, max_l: int, max_n: int = 5, cutoff: float = 5.0, smooth: bool = False):
        """
        Args:
            max_l: int, max order (excluding l)
            max_n: int, max number of roots used in each l
            cutoff: float, cutoff radius
        """
        self.max_l = max_l
        self.max_n = max_n
        self.cutoff = cutoff
        self.smooth = smooth
        if smooth:
            self.funcs = self._calculate_smooth_symbolic_funcs()
        else:
            self.funcs = self._calculate_symbolic_funcs()

        self.zeros = tf.cast(get_spherical_bessel_roots(), dtype=DataType.tf_float)

    @lru_cache(maxsize=128)
    def _calculate_symbolic_funcs(self) -> List:
        """
        Spherical basis functions based on Rayleigh formula. This function
        generates
        symbolic formula.

        Returns: list of symbolic functions

        """
        x = sympy.symbols("x")
        funcs = [sympy.expand_func(sympy.functions.special.bessel.jn(i, x)) for i in range(self.max_l + 1)]
        return [sympy.lambdify(x, func, "tensorflow") for func in funcs]

    @lru_cache(maxsize=128)
    def _calculate_smooth_symbolic_funcs(self) -> List:
        return _get_lambda_func(max_n=self.max_n, cutoff=self.cutoff)

    def __call__(self, r: tf.Tensor) -> tf.Tensor:
        """
        Args:
            r: tf.Tensor, distance Tensor, 1D


        Returns: [n, max_n * max_l] spherical Bessel function results

        """
        if self.smooth:
            return self._call_smooth_sbf(r)
        return self._call_sbf(r)

    def _call_smooth_sbf(self, r: tf.Tensor) -> tf.Tensor:
        results = [i(r) for i in self.funcs]
        return tf.transpose(tf.stack(results))

    def _call_sbf(self, r: tf.Tensor) -> tf.Tensor:
        roots = self.zeros[: self.max_l, : self.max_n]

        results = []
        factor = tf.cast(tf.sqrt(2.0 / self.cutoff**3), dtype=DataType.tf_float)
        for i in range(self.max_l):
            root = roots[i]
            func = self.funcs[i]
            func_add1 = self.funcs[i + 1]
            results.append(
                func(r[:, None] * root[None, :] / self.cutoff) * factor / tf.math.abs(func_add1(root[None, :]))
            )
        return tf.concat(results, axis=1)

    @staticmethod
    def rbf_j0(r: tf.Tensor, cutoff: float = 5.0, max_n: int = 3) -> tf.Tensor:
        """
        Spherical Bessel function of order 0, ensuring the function value
        vanishes at cutoff

        Args:
            r: tf.Tensor tensorflow tensors
            cutoff: float, the cutoff radius
            max_n: int max number of basis
        Returns: basis function expansion using first spherical Bessel function
        """
        n = tf.cast(tf.range(1, max_n + 1), dtype=DataType.tf_float)[None, :]
        r = r[:, None]
        return tf.math.sqrt(2.0 / cutoff) * tf.math.sin(n * pi / cutoff * r) / r


def _y00(theta, phi):
    r"""
    Spherical Harmonics with `l=m=0`

    ..math::
        Y_0^0 = \frac{1}{2} \sqrt{\frac{1}{\pi}}

    Args:
        theta: tf.Tensor, the azimuthal angle
        phi: tf.Tensor, the polar angle

    Returns: `Y_0^0` results

    """
    dtype = theta.dtype
    return 0.5 * tf.ones_like(theta) * tf.cast(tf.math.sqrt(1.0 / pi), dtype=dtype)


def _conjugate(x):
    return tf.math.conj(x)


class Gaussian:
    """
    Gaussian expansion function
    """

    def __init__(self, centers: Union[tf.Tensor, np.ndarray], width: float, **kwargs):
        """
        Args:
            centers (tf.Tensor or np.ndarray): Gaussian centers for the
                expansion
            width (float): Gaussian width
            **kwargs:
        """
        self.centers = np.array(centers)
        self.width = width

    def __call__(self, r: tf.Tensor) -> tf.Tensor:
        """
        Convert the radial distances into Gaussian functions
        Args:
            r (tf.Tensor): radial distances
        Returns: Gaussian expanded vectors

        """
        return tf.exp(-((r[:, None] - self.centers[None, :]) ** 2) / self.width**2)


class SphericalHarmonicsFunction:
    """
    Spherical Harmonics function
    """

    def __init__(self, max_l: int, use_phi: bool = True):
        """
        Args:
            max_l: int, max l (excluding l)
            use_phi: bool, whether to use the polar angle. If not,
                the function will compute `Y_l^0`
        """
        self.max_l = max_l
        self.use_phi = use_phi
        self.funcs = self._calculate_symbolic_funcs()

    def _calculate_symbolic_funcs(self):
        funcs = []
        theta, phi = sympy.symbols("theta phi")
        for lval in range(self.max_l):
            if self.use_phi:
                m_list = range(-lval, lval + 1)
            else:
                m_list = [0]
            for m in m_list:
                func = sympy.functions.special.spherical_harmonics.Znm(lval, m, theta, phi).expand(func=True)
                funcs.append(func)
        # replace all theta with cos(theta)
        costheta = sympy.symbols("costheta")
        funcs = [i.subs({theta: sympy.acos(costheta)}) for i in funcs]
        self.orig_funcs = [sympy.simplify(i).evalf() for i in funcs]
        results = [
            sympy.lambdify([costheta, phi], i, [{"conjugate": _conjugate}, "tensorflow"]) for i in self.orig_funcs
        ]
        results[0] = _y00
        return results

    def __call__(self, costheta, phi: Optional[tf.Tensor] = None):
        """
        Args:
            theta: tf.Tensor, the azimuthal angle
            phi: tf.Tensor, the polar angle

        Returns: [n, m] spherical harmonic results, where n is the number
            of angles. The column is arranged following
            `[Y_0^0, Y_1^{-1}, Y_1^{0}, Y_1^1, Y_2^{-2}, ...]`
        """
        costheta = tf.cast(costheta, dtype=tf.dtypes.complex64)
        phi = tf.cast(phi, dtype=tf.dtypes.complex64)
        results = tf.stack([func(costheta, phi) for func in self.funcs], axis=1)
        results = tf.cast(results, dtype=DataType.tf_float)
        return results


def _block_repeat(array, block_size, repeats):
    col_index = tf.range(tf.shape(array)[1])
    indices = []
    start = 0

    for i, b in enumerate(block_size):
        indices.append(tf.tile(col_index[start : start + b], [repeats[i]]))
        start += b
    indices = tf.concat(indices, axis=0)
    return tf.gather(array, indices, axis=1)


def combine_sbf_shf(sbf: tf.Tensor, shf: tf.Tensor, max_n: int, max_l: int, use_phi: bool):
    """
    Combine the spherical Bessel function and the spherical Harmonics function
    For the spherical Bessel function, the column is ordered by
        [n=[0, ..., max_n-1], n=[0, ..., max_n-1], ...], max_l blocks,

    For the spherical Harmonics function, the column is ordered by
        [m=[0], m=[-1, 0, 1], m=[-2, -1, 0, 1, 2], ...] max_l blocks, and each
        block has 2*l + 1
        if use_phi is False, then the columns become
        [m=[0], m=[0], ...] max_l columns

    Args:
        sbf: tf.Tensor spherical bessel function results
        shf: tf.Tensor spherical harmonics function results
        max_n: int, max number of n
        max_l: int, max number of l
        use_phi: whether to use phi
    Returns:
    """
    if tf.shape(sbf)[0] == 0:
        return sbf

    if not use_phi:
        repeats_sbf = [1] * max_l * max_n
        block_size = [1] * max_l
    else:
        # [1, 1, 1, ..., 1, 3, 3, 3, ..., 3, ...]
        repeats_sbf = np.repeat(2 * np.arange(max_l) + 1, repeats=max_n)
        # tf.repeat(2 * tf.range(max_l) + 1, repeats=max_n)
        block_size = 2 * np.arange(max_l) + 1
        # 2 * tf.range(max_l) + 1
    expanded_sbf = tf.repeat(sbf, repeats=repeats_sbf, axis=1)
    expanded_shf = _block_repeat(shf, block_size=block_size, repeats=[max_n] * max_l)
    shape = max_n * max_l
    if use_phi:
        shape *= max_l
    return tf.reshape(expanded_sbf * expanded_shf, [-1, shape])


def _sinc(x):
    return tf.math.sin(x) / x


@tf.function
def spherical_bessel_smooth(r, cutoff: float = 5.0, max_n: int = 10):
    """
    This is an orthogonal basis with first
    and second derivative at the cutoff
    equals to zero. The function was derived from the order 0 spherical Bessel
    function, and was expanded by the different zero roots

    Ref:
        https://arxiv.org/pdf/1907.02374.pdf

    Args:
        r: tf.Tensor distance tensor
        cutoff: float, cutoff radius
        max_n: int, max number of basis, expanded by the zero roots

    Returns: expanded spherical harmonics with derivatives smooth at boundary

    """
    n = tf.cast(tf.range(max_n), dtype=DataType.tf_float)[None, :]
    r = r[:, None]
    fnr = (
        (-1) ** n
        * tf.math.sqrt(2.0)
        * pi
        / cutoff**1.5
        * (n + 1)
        * (n + 2)
        / tf.math.sqrt(2 * n**2 + 6 * n + 5)
        * (_sinc(r * (n + 1) * pi / cutoff) + _sinc(r * (n + 2) * pi / cutoff))
    )
    en = n**2 * (n + 2) ** 2 / (4 * (n + 1) ** 4 + 1)
    dn = [tf.constant(1.0)]
    for i in range(1, max_n):
        dn.append(1 - en[0, i] / dn[-1])
    dn = tf.stack(dn)
    gn = [fnr[:, 0]]
    for i in range(1, max_n):
        gn.append(1 / tf.math.sqrt(dn[i]) * (fnr[:, i] + tf.sqrt(en[0, i] / dn[i - 1]) * gn[-1]))
    return tf.transpose(tf.stack(gn))


@lru_cache(maxsize=128)
def _get_lambda_func(max_n, cutoff: float = 5.0):
    r = sympy.symbols("r")
    d0 = 1.0
    en = []
    for i in range(max_n):
        en.append(i**2 * (i + 2) ** 2 / (4 * (i + 1) ** 4 + 1))

    dn = [d0]
    for i in range(1, max_n):
        dn.append(1 - en[i] / dn[-1])

    fnr = []
    for i in range(max_n):
        fnr.append(
            (-1) ** i
            * sympy.sqrt(2.0)
            * sympy.pi
            / cutoff**1.5
            * (i + 1)
            * (i + 2)
            / sympy.sqrt(1.0 * (i + 1) ** 2 + (i + 2) ** 2)
            * (
                sympy.sin(r * (i + 1) * sympy.pi / cutoff) / (r * (i + 1) * sympy.pi / cutoff)
                + sympy.sin(r * (i + 2) * sympy.pi / cutoff) / (r * (i + 2) * sympy.pi / cutoff)
            )
        )

    gnr = [fnr[0]]
    for i in range(1, max_n):
        gnr.append(1 / sympy.sqrt(dn[i]) * (fnr[i] + sympy.sqrt(en[i] / dn[i - 1]) * gnr[-1]))

    return [sympy.lambdify([r], sympy.simplify(i), "tensorflow") for i in gnr]



"""
atomic reference. Used for predicting extensive properties.
"""

import logging
from typing import List, Optional

import numpy as np
import tensorflow as tf
from ase import Atoms
from pymatgen.core import Element, Molecule, Structure

from m3gnet.config import DataType
from m3gnet.graph import Index
from m3gnet.utils import get_segment_indices_from_n, register

logger = logging.getLogger(__name__)


@register
class BaseAtomRef(tf.keras.layers.Layer):
    """
    Base AtomRef that predicts 0 correction
    """

    def call(self, graph: List, **kwargs):
        """

        Args:
            graph (list): list repr of a graph
            **kwargs:

        Returns: 0
        """
        return 0.0


@register
class AtomRef(BaseAtomRef):
    """
    Atom reference values. For example, if the average H energy is -20.0, and
    the average O energy is -10.0, the AtomRef predicts -20 * 2 + (-10.) =
    -50.0 for the atom reference energy for H2O
    """

    def __init__(
        self,
        property_per_element: Optional[np.ndarray] = None,
        max_z: int = 94,
        **kwargs,
    ):
        """
        Args:
            property_per_element (np.ndarray): element reference value
            max_z (int): maximum atomic number
            **kwargs:
        """
        super().__init__(**kwargs)
        if property_per_element is None:
            self.property_per_element = np.zeros(shape=(max_z + 1,))
        else:
            self.property_per_element = np.array(property_per_element).ravel()
        self.max_z = max_z

    def _get_feature_matrix(self, structs_or_graphs):
        n = len(structs_or_graphs)
        features = np.zeros(shape=(n, self.max_z + 1))
        for i, s in enumerate(structs_or_graphs):
            if isinstance(s, (Structure, Molecule)):
                atomic_numbers = [i.specie.Z for i in s.sites]
            elif isinstance(s, (list, np.ndarray)):
                atomic_numbers = s
            elif isinstance(s, Atoms):
                atomic_numbers = s.get_atomic_numbers()
            else:
                atomic_numbers = s.atoms[:, 0]
            features[i] = np.bincount(atomic_numbers, minlength=self.max_z + 1)
        return features

    def fit(self, structs_or_graphs, properties):
        """
        Fit the elemental reference values for the properties
        Args:
            structs_or_graphs (list): list of graphs or structures
            properties (np.ndarray): array of extensive properties
        Returns:
        """
        features = self._get_feature_matrix(structs_or_graphs)
        self.property_per_element = np.linalg.pinv(features.T.dot(features)).dot(features.T.dot(properties))
        string_prop = ""
        for i, j in enumerate(self.property_per_element):
            if abs(j) > 1e-5:
                string_prop += f"{str(Element.from_Z(i))}: {j:.5f}"
        logger.info("Property offset values: " + string_prop)
        return True

    def transform(self, structs_or_graphs, properties):
        """
        Correct the extensive properties by subtracting the atom reference
        values
        Args:
            structs_or_graphs (list): list of graphs or structures
            properties (np.ndarray): array of extensive properties
        Returns: corrected property values

        """
        properties = np.array(properties)
        atom_properties = self.predict_properties(structs_or_graphs)
        return properties - np.reshape(atom_properties, properties.shape)

    def inverse_transform(self, structs_or_graphs, properties):
        """
        Take the transformed values and get the original values
        Args:
            structs_or_graphs (list): list of graphs or structures
            properties (np.ndarray): array of extensive properties
        Returns: corrected property values

        """
        properties = np.array(properties)
        property_atoms = self.predict_properties(structs_or_graphs)
        final_properties = properties + np.reshape(property_atoms, properties.shape)
        return final_properties

    def predict_properties(self, structs_or_graphs):
        """
        Args:
            structs_or_graphs (list): calculate the atom summed property
                values
        Returns:

        """
        if not isinstance(structs_or_graphs, list):
            structs_or_graphs = [structs_or_graphs]
        features = self._get_feature_matrix(structs_or_graphs)
        return features.dot(self.property_per_element)

    def call(self, graph: List, **kwargs):
        """
        Args:
            graph (list): a list repr of a graph
            **kwargs:
        Returns:
        """
        atomic_numbers = graph[Index.ATOMS][:, 0]
        atom_energies = tf.gather(tf.cast(self.property_per_element, DataType.tf_float), atomic_numbers)
        res = tf.math.segment_sum(atom_energies, get_segment_indices_from_n(graph[Index.N_ATOMS]))
        return tf.reshape(res, (-1, 1))

    def set_property_per_element(self, property_per_element):
        """
        Set the property per atom value
        Args:
            property_per_element (np.ndarray): array of elemental properties,
                the i-th row is the elemental value for atomic number i.
        Returns:
        """
        self.property_per_element = property_per_element

    def get_config(self):
        """
        Get dict config for serialization
        Returns (dict):
        """
        config = super().get_config()
        config.update(**{"property_per_element": self.property_per_element, "max_z": self.max_z})
        return config



# -*- coding: utf-8 -*-
import unittest

import numpy as np

from m3gnet.graph import Index
from m3gnet.layers import AtomReduceState


class TestAgg(unittest.TestCase):
    def test_agg(self):
        graph = [[] for _ in range(13)]
        graph[Index.ATOMS] = np.random.normal(size=(10, 3))
        graph[Index.N_ATOMS] = np.array([5, 5])
        graph[Index.STATES] = np.random.normal(size=(2, 5))
        ars = AtomReduceState(method="mean")
        res = ars(graph)
        self.assertTrue(np.allclose(res[0], np.mean(graph[Index.ATOMS][:5], axis=0), atol=1e-5))
        self.assertTrue(np.allclose(res[1], np.mean(graph[Index.ATOMS][5:], axis=0), atol=1e-5))


if __name__ == "__main__":
    unittest.main()



"""
Common metrics used in M3GNet
"""
from typing import Callable

import tensorflow as tf

METRICS = {"rmse": tf.keras.metrics.RootMeanSquaredError()}

REVERSE_NAME_MAPPING = {"mean_absolute_error": "mae"}

MONITOR_MAPPING = {"val_AUC": "max", "val_mae": "min", "val_loss": "min"}


def _get_metric(metric):
    try:
        return tf.keras.metrics.get(metric)
    except ValueError:
        if isinstance(metric, Callable):
            return metric

        if metric.lower() not in METRICS:
            raise ValueError("metric not found")
        return METRICS[metric.lower()]


def _get_metric_string(metric):
    if isinstance(metric, tf.keras.metrics.Metric):
        return metric.__class__.__name__
    name = metric.__name__
    if name in REVERSE_NAME_MAPPING:
        return REVERSE_NAME_MAPPING[name]
    return name



"""
Graph layers
"""
from ._aggregate import AtomReduceState
from ._atom import AtomNetwork, GatedAtomUpdate
from ._atom_ref import AtomRef, BaseAtomRef
from ._base import GraphUpdate, GraphUpdateFunc
from ._basis import RadialBasisFunctions
from ._bond import BondNetwork, ConcatAtoms, PairRadialBasisExpansion, ThreeDInteraction
from ._core import METHOD_MAPPING, MLP, Embedding, GatedMLP, Pipe
from ._cutoff import cosine, polynomial
from ._gn import GraphFeaturizer, GraphFieldEmbedding, GraphNetworkLayer
from ._readout import (
    MultiFieldReadout,
    ReadOut,
    ReduceReadOut,
    Set2Set,
    WeightedReadout,
)
from ._state import ConcatBondAtomState, StateNetwork
from ._three_body import SphericalBesselWithHarmonics
from ._two_body import PairDistance, PairVector

__all__ = [
    "GraphUpdate",
    "GraphUpdateFunc",
    "METHOD_MAPPING",
    "Pipe",
    "MLP",
    "GatedMLP",
    "Embedding",
    "RadialBasisFunctions",
    "polynomial",
    "cosine",
    "PairDistance",
    "PairVector",
    "AtomReduceState",
    "SphericalBesselWithHarmonics",
    "BondNetwork",
    "PairRadialBasisExpansion",
    "ConcatAtoms",
    "ThreeDInteraction",
    "AtomNetwork",
    "GatedAtomUpdate",
    "StateNetwork",
    "ConcatBondAtomState",
    "ReadOut",
    "ReduceReadOut",
    "Set2Set",
    "MultiFieldReadout",
    "WeightedReadout",
    "GraphNetworkLayer",
    "GraphFeaturizer",
    "GraphFieldEmbedding",
    "AtomRef",
    "BaseAtomRef",
]



"""
Cutoff functions
"""

import numpy as np
import tensorflow as tf


def polynomial(r: tf.Tensor, cutoff: float) -> tf.Tensor:
    """
    Polynomial cutoff function
    Args:
        r (tf.Tensor): radius distance tensor
        cutoff (float): cutoff distance

    Returns: polynomial cutoff functions

    """
    ratio = r / cutoff
    return tf.where(r <= cutoff, 1 - 6 * ratio**5 + 15 * ratio**4 - 10 * ratio**3, 0.0)


def cosine(r: tf.Tensor, cutoff: float) -> tf.Tensor:
    """
    Cosine cutoff function
    Args:
        r (tf.Tensor): radius distance tensor
        cutoff (float): cutoff distance

    Returns: cosine cutoff functions

    """
    return tf.where(r <= cutoff, tf.math.cos(np.pi * r / cutoff), 0.0)


CUTOFF_MAPPING = {"polynomial": polynomial, "cosine": cosine}



"""
state networks

The global state networks take a graph as inputs and calculate the updated
global attributes
"""

from typing import Callable, List

import tensorflow as tf

from m3gnet.graph import Index
from m3gnet.utils import register

from ._base import GraphUpdate


@register
class StateNetwork(GraphUpdate):
    """
    Edge network that takes a graph as input and then calculate new
    bond attributes
    """

    def update_states(self, graph: List) -> tf.Tensor:
        """
        Calculate the new state attributes
        Args:
            graph (list): list repr of a MaterialGraph

        Returns: tf.Tensor
        """
        return graph[Index.STATES]

    def call(self, graph: List, **kwargs) -> List:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns: new graph in list repr

        """
        graph = graph[:]
        states = self.update_states(graph)
        graph[Index.STATES] = states
        return graph


@register
class ConcatBondAtomState(StateNetwork):
    r"""
    u^\prime = Update(\bar e^\prime⊕\bar v^\prime⊕u)
    """

    def __init__(self, update_func: Callable, bond_agg_func: Callable = None, atom_agg_func: Callable = None, **kwargs):
        """
        Args:
            update_func (callable): the core update function
            bond_agg_func (callable): function to aggregate bond to state
            atom_agg_func (callable): function to aggregate atom to state
            **kwargs:
        """
        self.update_func = update_func
        self.bond_agg_func = bond_agg_func
        self.atom_agg_func = atom_agg_func
        super().__init__(**kwargs)

    def update_states(self, graph: List) -> tf.Tensor:
        """
        Args:
            graph (list): list repr of a MaterialGraph
        Returns:
        """
        info_list = [graph[Index.STATES]]
        if self.bond_agg_func is not None:
            bond_agg = self.bond_agg_func(graph)
            info_list.append(bond_agg)
        if self.atom_agg_func is not None:
            atom_agg = self.atom_agg_func(graph)
            info_list.append(atom_agg)

        if len(info_list) > 1:
            concat = tf.concat(info_list, axis=-1)
        else:
            concat = info_list[0]
        return self.update_func(concat)

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update(
            {
                "update_func": self.update_func,
                "bond_agg_func": self.bond_agg_func,
                "atom_agg_func": self.atom_agg_func,
            }
        )
        return config



"""
Edge networks.

The edge networks take a graph as inputs and calculate the updated
edge attributes
"""

from typing import List

import tensorflow as tf

from m3gnet.graph import Index
from m3gnet.utils import broadcast_states_to_bonds, register

from ._base import GraphUpdate
from ._core import GatedMLP


@register
class AtomNetwork(GraphUpdate):
    """
    Atom network that takes a graph as input and then calculate new
    atom attributes
    """

    def update_atoms(self, graph: List) -> tf.Tensor:
        """
        Take a graph input and calculate the updated atom attributes
        Args:
            graph (list): list repr of a MaterialGraph

        Returns: tf.Tensor

        """
        return graph[Index.ATOMS]

    def call(self, graph: List, **kwargs) -> List:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns: a new MaterialGraph in list repr

        """
        graph = graph[:]
        atoms = self.update_atoms(graph)
        graph[Index.ATOMS] = atoms
        return graph


@register
class GatedAtomUpdate(AtomNetwork):
    """
    Take the neighbor atom attributes and bond attributes, update them to the
    center atom
    """

    def __init__(self, neurons: List[int], activation: str = "swish", **kwargs):
        """
        Args:
            neurons (list): number of neurons in each layer
            activation (str): activation function
            **kwargs:
        """
        super().__init__(**kwargs)
        self.neurons = neurons
        self.bond_dense = GatedMLP(neurons=neurons, activations=[activation] * len(neurons))
        self.weight_update = tf.keras.layers.Dense(neurons[-1], activation=activation, use_bias=False)
        self.activation = activation

    def _get_reduced(self, graph: List):
        atoms_left = tf.gather(graph[Index.ATOMS], graph[Index.BOND_ATOM_INDICES][:, 0])
        atoms_right = tf.gather(graph[Index.ATOMS], graph[Index.BOND_ATOM_INDICES][:, 1])

        atoms = [atoms_left, atoms_right, graph[Index.BONDS]]

        if graph[Index.STATES] is not None:
            atoms.append(broadcast_states_to_bonds(graph))

        reduced = tf.concat(atoms, axis=-1)
        reduced = self.bond_dense(reduced) * self.weight_update(graph[Index.BOND_WEIGHTS])
        return reduced

    def update_atoms(self, graph: List) -> tf.Tensor:
        """
        Update atom attributes
        Args:
            graph (list): list repr of MaterialGraph
        Returns: tf.Tensor

        """
        reduced = self._get_reduced(graph)
        c = tf.math.unsorted_segment_sum(
            reduced,
            graph[Index.BOND_ATOM_INDICES][:, 0],
            num_segments=tf.shape(graph[Index.ATOMS])[0],
        )
        return graph[Index.ATOMS] + c

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update(neurons=self.neurons, activation=self.activation)
        return config



# -*- coding: utf-8 -*-
import unittest

import numpy as np
import tensorflow as tf

from m3gnet.utils import (
    Gaussian,
    SphericalBesselFunction,
    SphericalHarmonicsFunction,
    combine_sbf_shf,
    get_spherical_bessel_roots,
    spherical_bessel_roots,
    spherical_bessel_smooth,
)


class TestMath(unittest.TestCase):
    def test_spherical_bessel_roots(self):
        roots = spherical_bessel_roots(max_l=1, max_n=5)
        roots2 = get_spherical_bessel_roots()
        self.assertTrue(np.allclose(roots2[0, :5], roots.ravel()))

    def test_gaussian(self):
        centers = np.linspace(1.0, 5.0, 10)
        width = 0.5
        gf = Gaussian(centers=centers, width=width)
        r = np.linspace(1.0, 4.0, 10)
        self.assertTupleEqual(gf(r).numpy().shape, (10, 10))

    def test_spherical_bessel_harmonics_function(self):
        r = tf.random.normal(shape=(10,))
        sbf = SphericalBesselFunction(max_l=3, cutoff=5.0, max_n=3, smooth=False)
        res = sbf(r)
        res2 = sbf.rbf_j0(r, cutoff=5.0, max_n=3)
        self.assertTrue(np.allclose(res[:, :3].numpy().ravel(), res2.numpy().ravel()))

        self.assertTrue(res.numpy().shape == (10, 9))

        sbf2 = SphericalBesselFunction(max_l=3, cutoff=5.0, max_n=3, smooth=True)

        res2 = sbf2(r)
        self.assertTupleEqual(tuple(res2.shape), (10, 3))

        shf = SphericalHarmonicsFunction(max_l=3, use_phi=True)
        res_shf = shf(costheta=np.linspace(-1, 1, 10), phi=np.linspace(0, 2 * np.pi, 10))

        self.assertTrue(res_shf.numpy().shape == (10, 9))
        combined = combine_sbf_shf(res, res_shf, max_n=3, max_l=3, use_phi=True)

        self.assertTrue(combined.shape == (10, 27))

        res_shf2 = SphericalHarmonicsFunction(max_l=3, use_phi=False)(
            costheta=np.linspace(-1, 1, 10), phi=np.linspace(0, 2 * np.pi, 10)
        )
        combined = combine_sbf_shf(res, res_shf2, max_n=3, max_l=3, use_phi=False)

        self.assertTrue(combined.shape == (10, 9))
        rdf = spherical_bessel_smooth(r, cutoff=5.0, max_n=3)
        self.assertTrue(rdf.numpy().shape == (10, 3))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from pymatgen.core import Lattice, Molecule, Structure

from m3gnet.graph import RadiusCutoffGraphConverter, tf_compute_distance_angle
from m3gnet.layers import PairDistance, PairVector, SphericalBesselWithHarmonics


class TestTwoBody(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])
        mol = Molecule(["C", "O"], [[0, 0, 0], [1.1, 0, 0]])
        rc = RadiusCutoffGraphConverter(cutoff=4.0, has_threebody=True, threebody_cutoff=4.0)
        cls.struct_graph = rc.convert(s).as_list()
        cls.mol_graph = rc.convert(mol).as_list()

    def test_twobody(self):
        pv = PairVector()
        self.assertTrue(pv(self.struct_graph).shape == (28, 3))
        self.assertTrue(pv(self.mol_graph).shape == (2, 3))

        pd = PairDistance()
        self.assertTrue(pd(self.struct_graph).shape == (28,))
        self.assertTrue(np.allclose(pd(self.mol_graph), np.array([1.1, 1.1])))

    def test_threebody(self):
        sph = SphericalBesselWithHarmonics(max_n=3, max_l=3, cutoff=4.0)
        sarray = sph(tf_compute_distance_angle(self.struct_graph))
        self.assertTrue(sarray.shape[1] == 9)
        marray = sph(tf_compute_distance_angle(self.mol_graph))
        self.assertTrue(marray.shape[1] == 9)


if __name__ == "__main__":
    unittest.main()



"""
Graph types.
"""

from dataclasses import dataclass, replace
from typing import ClassVar, List, Optional, Sequence, Union

import numpy as np
import tensorflow as tf

from m3gnet.utils import check_array_equal, check_shape_consistency


def _to_numpy(x):
    if isinstance(x, Sequence):
        return np.array(x)
    if isinstance(x, tf.Tensor):
        return x.numpy()
    if isinstance(x, np.ndarray):
        return x
    raise ValueError("Cannot convert to numpy")


#  graph features
ATOMS = "atoms"
BONDS = "bonds"
STATES = "states"
GRAPH_FEATURES = (ATOMS, BONDS, STATES)

# atom positions, some pretrained may need explicit
# atom positions
ATOM_POSITIONS = "atom_positions"
BOND_WEIGHTS = "bond_weights"

# atom indices that define bonds
BOND_ATOM_INDICES = "bond_atom_indices"

# number of atoms and graphs in each structure
N_ATOMS = "n_atoms"
N_BONDS = "n_bonds"
ATOM_BOND_NUMBERS = (N_ATOMS, N_BONDS)

# periodic boundary fields
PBC_OFFSETS = "pbc_offsets"
LATTICES = "lattices"
PBC_FIELDS = (PBC_OFFSETS, LATTICES)

# threebody information
TRIPLE_BOND_INDICES = "triple_bond_indices"
TRIPLE_BOND_LENGTHS = "triple_bond_lengths"
THETA = "theta"
PHI = "phi"
N_TRIPLE_IJ = "n_triple_ij"
N_TRIPLE_I = "n_triple_i"
N_TRIPLE_S = "n_triple_s"


ALL_FIELDS = (
    ATOMS,
    BONDS,
    STATES,
    ATOM_POSITIONS,
    BOND_ATOM_INDICES,
    PBC_OFFSETS,
    N_ATOMS,
    N_BONDS,
    BOND_WEIGHTS,
    LATTICES,
    TRIPLE_BOND_INDICES,
    TRIPLE_BOND_LENGTHS,
    THETA,
    PHI,
    N_TRIPLE_IJ,
    N_TRIPLE_I,
    N_TRIPLE_S,
)


GRAPH_INDEX = {i: j for j, i in enumerate(ALL_FIELDS)}


class Index:
    """
    Get integer indices for each field in MaterialGraph
    """

    ATOMS = 0
    BONDS = 1
    STATES = 2
    ATOM_POSITIONS = 3
    BOND_ATOM_INDICES = 4
    PBC_OFFSETS = 5
    N_ATOMS = 6
    N_BONDS = 7
    BOND_WEIGHTS = 8
    LATTICES = 9
    TRIPLE_BOND_INDICES = 10
    TRIPLE_BOND_LENGTHS = 11
    THETA = 12
    PHI = 13
    N_TRIPLE_IJ = 14
    N_TRIPLE_I = 15
    N_TRIPLE_S = 16


class AttributeUpdateMixin:
    """
    Mixin class for updating the fields of the graphs and checking
    graph consistency
    """

    def _check_graph(self) -> None:
        """
        Check if the graph is valid
        """
        pass

    def replace(self, **kwargs):
        """
        Replace a graph field
        Args:
            **kwargs: dictionary for replacements

        Returns:

        """
        return replace(self, **kwargs)  # noqa

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        for i, j in self.__dict__.items():
            is_equal = check_array_equal(j, other.__dict__[i])
            if not is_equal:
                print(
                    f"{i} is different, one is ",
                    j,
                    "and the other is",
                    other.__dict__[i],
                )
                return False
        return True

    def _check_shapes(self, shape_dict):
        checks = [check_shape_consistency(getattr(self, i), j) for i, j in shape_dict.items()]
        names = list(shape_dict.keys())
        error_msg = ""
        for check, name in zip(checks, names):
            if not check:
                error_msg += f"{name} shape is {getattr(self, name).shape}, "
                error_msg += f"while expecting {shape_dict[name]}"

        if len(error_msg) > 0:
            raise RuntimeError("The data shapes do not comply\n" + error_msg)


def _check_n(fields, name, dim=0):
    """
    Check the field consistency
    Args:
        fields (list): a list of fields
        name (str): name to use when raising error
        dim (int): dimension for comparing

    Returns: int or None
    """
    n_candidates = []
    for i in fields:
        if i is not None:
            n_candidates.append(i.shape[dim])
    if len(n_candidates) > 0:
        if len(list(set(n_candidates))) > 1:
            raise ValueError(f"{name} inconsistent")
        return n_candidates[0]
    return None


@dataclass(eq=False)
class MaterialGraph(AttributeUpdateMixin):
    """
    Material graph by an edge list graph representation.
    In this representation, no padding of atoms or bonds are needed.

    Full available data names are the following, assuming the
    number of atoms is `n_atom`, number of bonds is `n_bond` and
    number of structure is `n_struct`.

    The attributes are as follows

        - `atoms` (np.ndarray): atom attributes with shape [n_atom, None]
        - `bonds` (np.ndarray): bond attributes with shape [n_bond, None]
            states (np.ndarray): state attributes with shape [n_struct, None]
            bond_atom_indices (np.ndarray): int indices for pairs of atoms,
            shape [n_bond, 2]
        - `n_atoms` (np.ndarray): int array, number of atoms in each structure,
            [n_struct]
        - `n_bonds` (np.ndarray): int array, number of bonds in each structure,
            [n_struct]
        - `atom_positions` (np.ndarray): float type, atom position array, shape
            [n_atom, 3]
        - `bond_weights` (np.ndarray): float type, bond lengths, shape [n_bond]
        - `pbc_offsets` (np.ndarray): int type, periodic boundary offset
             vectors, shape [n_bond, 3].
        - `lattices` (np.ndarray): float type, lattice matrices for all
            structures, shape [n_struct, 3, 3]
        - `triple_bond_indices` (np.ndarray): int type, the bond indices for
            forming triple bonds
        - `triple_bond_lengths` (np.ndarray): float type, the triple bond
            lengths
        - `theta` (np.ndarray): float type,the azimuthal angle of the triple
            bond
        - `phi` (np.ndarray): float type, the polar angle of the triple bond
        - `n_triple_ij` (np.ndarray): int type, the number of triple bonds for
            each bond
        -  `n_triple_i` (np.ndarray): int type, the number of triple bonds for
            each atom
        - `n_triple_s` (np.ndarray): int type, the number of triple bonds
            for each structure
    """

    atoms: Optional[Union[np.ndarray, tf.Tensor]] = None  # [n_atom, None]
    bonds: Optional[Union[np.ndarray, tf.Tensor]] = None  # [n_bond, None]
    states: Optional[Union[np.ndarray, tf.Tensor]] = None  # [n_struct, None]
    # [n_bond, 2], int
    bond_atom_indices: Optional[Union[np.ndarray, tf.Tensor]] = None
    n_atoms: Optional[Union[np.ndarray, tf.Tensor]] = None  # [n_struct], int
    n_bonds: Optional[Union[np.ndarray, tf.Tensor]] = None  # [n_struct], int

    # the following are optional
    # [n_atom, 3], float
    atom_positions: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_bond], float
    bond_weights: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_bond, 3], int
    pbc_offsets: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_struct, 3, 3], float
    lattices: Optional[Union[np.ndarray, tf.Tensor]] = None

    # The following are the three body information
    # [n_triple, 2], int
    triple_bond_indices: Optional[Union[np.ndarray, tf.Tensor]] = None

    # [n_triple, ], float
    triple_bond_lengths: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_triple, ], float
    theta: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_triple, ], float
    phi: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_bond, ], int
    n_triple_ij: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_atom, ], int
    n_triple_i: Optional[Union[np.ndarray, tf.Tensor]] = None
    # [n_struct, ], int
    n_triple_s: Optional[Union[np.ndarray, tf.Tensor]] = None

    # the following are class attributes
    base_attributes: ClassVar[List[str]] = [
        "atoms",
        "bonds",
        "bond_atom_indices",
        "n_atoms",
        "n_bonds",
    ]

    graph_with_position_attributes: ClassVar[List[str]] = [
        "atoms",
        "bond_atom_indices",
        "atom_positions",
        "bond_weights",
        "n_atoms",
        "n_bonds",
        "pbc_offsets",
        "lattices",
    ]

    three_body_attributes: ClassVar[List[str]] = [
        "triple_bond_indices",
        "triple_bond_lengths",
        "theta",
        "phi",
        "n_triple_ij",
        "n_triple_i",
        "n_triple_s",
    ]

    def __repr__(self):
        attributes = ALL_FIELDS
        shapes = []
        for attr in attributes:
            if getattr(self, attr) is None:
                continue
            shapes.append(attr + " " + str(getattr(self, attr).dtype)[9:-2] + f" {str(getattr(self, attr).shape)}")
        string = "\n" + "\n".join(shapes)
        return "<MaterialGraph with the following data shapes: " + string + ">"

    def as_tf(self) -> "MaterialGraph":
        """
        Convert each field to tensorflow tensors
        Returns: tf.Tensor
        """
        d = {i: _maybe_tensor(getattr(self, i)) for i in ALL_FIELDS}
        mg = MaterialGraph(**d)  # type: ignore
        return mg

    @property
    def has_threebody(self) -> bool:
        """
        Whether the graph has threebody indices
        Returns: boolean value indicator

        """
        if self.n_triple_ij is not None:
            return True
        return False

    @property
    def n_atom(self) -> Optional[int]:
        """
        number of atoms in the graph
        Returns: int or None
        """
        return None if self.atoms is None else self.atoms.shape[0]

    @property
    def n_bond(self) -> Optional[int]:
        """
        number of bonds
        Returns: int or None
        """
        return None if self.bonds is None else self.bonds.shape[0]

    @property
    def n_struct(self) -> Optional[int]:
        """
        number of structures
        Returns: int or None
        """
        if self.states is not None:
            return self.states.shape[0]
        if self.n_atoms is not None:
            return len(self.n_atoms)
        return None

    def as_list(self) -> List:
        """
        Convert the MaterialGraph to list representation
        Returns:

        """
        return [getattr(self, i) for i in ALL_FIELDS]

    @classmethod
    def from_list(cls, graph_list):
        """
        Construct the MaterialGraph object from a list representation
        Args:
            graph_list (list): list representation of a MaterialGraph

        Returns: MaterialGraph instance

        """
        graph_list = [np.array(i) if i is not None else i for i in graph_list]
        return cls(**dict(zip(ALL_FIELDS, graph_list)))

    def copy(self) -> "MaterialGraph":
        """
        Copy the MaterialGraph to a new one
        Returns: MaterialGraph
        """

        graph_list = self.as_list()
        graph_list_copy = [i.copy() if i is not None else None for i in graph_list]
        return MaterialGraph.from_list(graph_list_copy)


def _maybe_tensor(inp):
    """
    If the input is not a tensor, convert it to tensor
    Args:
        inp (object): input tensor or array or list

    Returns: tf.Tensor

    """
    if inp is not None:
        return tf.constant(inp)
    return None



import unittest

import numpy as np

from m3gnet.utils import check_array_equal, check_shape_consistency, reshape_array


class TestGeneralUtils(unittest.TestCase):
    def test_check_array_equal(self):
        self.assertTrue(check_array_equal(None, None))
        self.assertFalse(check_array_equal(None, np.array([1, 2, 3])))
        self.assertFalse(check_array_equal(np.array([1, 2, 3]), None))
        self.assertTrue(check_array_equal(np.array([1, 2, 3]), np.array([1, 2, 3])))

    def test_check_shape_consistency(self):
        self.assertTrue(check_shape_consistency(np.random.normal(size=(1, 2, 3)), [1, 2, None]))
        self.assertTrue(check_shape_consistency(np.random.normal(size=(1, 2, 3)), [1, 2, 3]))
        self.assertTrue(check_shape_consistency(np.random.normal(size=(1, 2, 3)), [1, None]))

    def test_reshape(self):
        self.assertRaises(ValueError, reshape_array, np.array([1, 2, 3]), [2, None])
        self.assertEqual(reshape_array(np.array([1, 2, 3]), [3, None]).shape, (3, 1))


if __name__ == "__main__":
    unittest.main()



"""
Base model
"""

import math
from abc import ABC
from typing import List, Union

import tensorflow as tf
from pymatgen.core import Molecule, Structure

from m3gnet.config import DataType
from m3gnet.graph import Index, MaterialGraph, assemble_material_graph
from m3gnet.type import StructureOrMolecule
from m3gnet.utils import register, repeat_with_n


@register
class GraphModelMixin(tf.keras.layers.Layer):
    """
    GraphModelMixin adds the following functionality to a graph model
        - predict_structure
        - predict_structures
        - predict_graph
        - predict_graphs
    """

    def predict_structure(self, structure: StructureOrMolecule) -> tf.Tensor:
        """
        predict properties from structure
        Args:
            structure (StructureOrMolecule): a pymatgen Structure/Molecule,
                or ase.Atoms
        Returns: predicted values
        """
        return self.predict_graph(self.graph_converter(structure))

    def predict_structures(self, structures: List[StructureOrMolecule], batch_size: int = 128) -> tf.Tensor:
        """
        predict properties from structures
        Args:
            structures (list[StructureOrMolecule]): a list of structures
            batch_size (int): batch size for the prediction
        Returns: predicted values
        """
        graph_list = [self.graph_converter(i) for i in structures]
        return self.predict_graphs(graph_list, batch_size)

    def predict_graph(self, graph: Union[MaterialGraph, List]) -> tf.Tensor:
        """
        predict properties from a graph
        Args:
            graph (Union[MaterialGraph, List]): a material graph, either in
                object or list repr
        Returns: predicted property
        """
        if isinstance(graph, MaterialGraph):
            graph = graph.as_list()
        return self.call(graph)

    def predict_graphs(self, graph_list: List[Union[MaterialGraph, List]], batch_size: int = 128) -> tf.Tensor:
        """
        predict properties from graphs
        Args:
            graph_list (List[Union[MaterialGraph, List]]): a list of material
                graph, either in object or list repr
            batch_size (int): batch size
        Returns: predicted properties
        """
        n = len(graph_list)
        use_graph = bool(isinstance(graph_list[0], MaterialGraph))
        n_steps = math.ceil(n / batch_size)
        predicted = []
        for i in range(n_steps):
            graphs = graph_list[batch_size * i : batch_size * (i + 1)]
            graph = assemble_material_graph(graphs)  # type: ignore
            if use_graph:
                results = self.call(graph.as_list())
            else:
                results = self.call(graph)
            predicted.append(results)
        return tf.concat(predicted, axis=0)


@register
class BasePotential(tf.keras.Model, ABC):
    """
    Potential abstract class
    """

    def get_energies(self, graph: List):
        """
        Compute the energy of a MaterialGraph
        Args:
            graph: List, a graph from structure
        Returns: energy values, size [Ns]
        """
        return 0.0

    def get_forces(self, graph: List):
        """
        Compute forces of a graph given the atom positions
        Args:
            graph: List, a graph from structure

        Returns: forces [Na, 3]
        """
        return self.get_efs(graph)[1]

    def get_stresses(self, graph: List):
        """
        Compute stress of a graph given the atom positions
        Args:
            graph: List, a graph from structure
        Returns: stresses [Ns, 3, 3]
        """
        return self.get_efs(graph)[2]

    def get_ef(self, obj: Union[StructureOrMolecule, MaterialGraph, List]) -> tuple:
        """
        get energy and force from a Structure, a graph or a list repr of a
        graph
        Args:
            obj (Union[StructureOrMolecule, MaterialGraph, List]): a structure,
                material graph or list repr of a graph
        Returns:
        """
        return self.get_efs(obj, include_stresses=False)

    @tf.function(experimental_relax_shapes=True)
    def get_ef_tensor(self, graph: List[tf.Tensor]) -> tuple:
        """
        get energy and force from a list repr of graph
        Args:
            graph (List[tf.Tensor]: a list repr of a graph
        Returns:
        """
        return self.get_efs_tensor(graph, include_stresses=False)

    def get_efs(
        self,
        obj: Union[StructureOrMolecule, MaterialGraph, List],
        include_stresses: bool = True,
    ):
        """
        get energy and force from a Structure, a graph or a list repr of a
        graph
        Args:
            obj (Union[StructureOrMolecule, MaterialGraph, List]): a structure,
                material graph or list repr of a graph
            include_stresses (bool): whether to include stress
        Returns:
        """
        if isinstance(obj, Structure):
            obj = self.model.graph_converter(obj)
        if isinstance(obj, MaterialGraph):
            obj = obj.as_tf().as_list()
        return self.get_efs_tensor(obj, include_stresses=include_stresses)

    @tf.function(experimental_relax_shapes=True)
    def get_efs_tensor(self, graph: List[tf.Tensor], include_stresses: bool = True) -> tuple:
        """
        get energy and force from a list repr of a
        graph
        Args:
            graph (List[tf.Tensor]): a list repr of a graph
            include_stresses (bool): whether to include stress
        Returns:
        """
        with tf.GradientTape() as tape:
            tape.watch(graph[Index.ATOM_POSITIONS])
            if include_stresses:
                graph = graph[:]
                strain = tf.zeros_like(graph[Index.LATTICES])
                tape.watch(strain)
                graph[Index.LATTICES] = tf.matmul(graph[Index.LATTICES], (tf.eye(3)[None, ...] + strain))

                strain_augment = repeat_with_n(strain, graph[Index.N_ATOMS])
                graph[Index.ATOM_POSITIONS] = tf.keras.backend.batch_dot(
                    graph[Index.ATOM_POSITIONS], (tf.eye(3)[None, ...] + strain_augment)
                )
                volume = tf.linalg.det(graph[Index.LATTICES])
            energies = self.get_energies(graph)
            derivatives = {"forces": graph[Index.ATOM_POSITIONS]}
            if include_stresses:
                derivatives["stresses"] = strain  # type: ignore

            derivatives = tape.gradient(energies, derivatives)

        forces = -derivatives["forces"]
        forces = tf.cast(tf.convert_to_tensor(forces), DataType.tf_float)
        results: tuple = (energies, forces)
        # eV/A^3 to GPa
        if include_stresses:
            stresses = 1 / volume[:, None, None] * derivatives["stresses"] * 160.21766208
            stresses = tf.cast(tf.convert_to_tensor(stresses), DataType.tf_float)
            results += (stresses,)
        return results

    def call(
        self,
        graph: Union[MaterialGraph, Structure, Molecule, List],
        include_forces: bool = True,
        include_stresses: bool = True,
    ):
        """
        Apply the potential to a graph
        Args:
            graph (Union[MaterialGraph, Structure, Molecule, List]): a
                structure, molecule, graph or a list repr of a graph
            include_forces (bool): whether to include forces as outputs
            include_stresses (bool: whether to include stresses as outputs
        Returns: energy [forces, stress]
        """
        if isinstance(graph, (Structure, Molecule)):
            graph = self.graph_converter.convert(graph)
        efs = self.get_efs(graph, include_stresses=include_stresses)
        results: tuple = (efs[0],)
        if include_forces:
            results += (efs[1],)
        if include_stresses:
            results += (efs[2],)
        if len(results) == 1:
            return results[0]
        return results


@register
class Potential(BasePotential):
    """
    Defines the Potential class. New potential should subclass from this class
    and define the "get_energies" method
    """

    def __init__(self, model: tf.keras.layers.Layer, **kwargs):
        """
        Args:
            model (tf.keras.layers.Layer): a callable model that predict values
                from a graph
            **kwargs:
        """
        super().__init__(**kwargs)
        self.model = model
        self.graph_converter = model.graph_converter

    @tf.function(experimental_relax_shapes=True)
    def get_energies(self, graph: List) -> tf.Tensor:
        """
        get energies from a list repr of a graph
        Args:
            graph (List): list repr of a graph
        Returns:
        """
        return self.model(graph)



"""
Calculate distance from atom positions and indices
"""
from typing import List

import tensorflow as tf

from m3gnet.graph import get_pair_vector_from_graph
from m3gnet.utils import get_length, register


@register
class PairDistance(tf.keras.layers.Layer):
    """
    Compute pair distances from atom positions, bond indices, lattices and
    periodic offsets.
    """

    def call(self, graph: List, **kwargs):
        """
        Calculate the pair distance from a MaterialGraph.
        Args:
            graph (list): A list representation of a MaterialGraph object
            **kwargs:

        Returns: tf.Tensor distance tensor

        """
        pair_vectors = get_pair_vector_from_graph(graph)
        return get_length(pair_vectors)


@register
class PairVector(tf.keras.layers.Layer):
    """
    Compute pair atom distance vectors from graph
    """

    def call(self, graph: List, **kwargs):
        """
        Calculate the pair vector distance from a MaterialGraph.
        Args:
            graph (List): A MaterialGraph object
            **kwargs:

        Returns: tf.Tensor distance vector tensor
        """
        return get_pair_vector_from_graph(graph)



"""
Classes to convert a structure into a graph
"""
import logging
from abc import abstractmethod
from typing import Dict, List, Optional

import numpy as np
import tensorflow as tf
from ase import Atoms
from pymatgen.core import Structure

from m3gnet.config import DataType
from m3gnet.type import StructureOrMolecule
from m3gnet.utils import register, reshape_array

from ._batch import assemble_material_graph
from ._compute import include_threebody_indices
from ._structure import get_fixed_radius_bonding
from ._types import MaterialGraph

logger = logging.getLogger(__name__)


@register
class BaseGraphConverter(tf.keras.layers.Layer):
    """
    Basic graph converter that uses the atomic number as the node feature
    """

    def __init__(self, default_states=None, **kwargs):
        """
        Args:
            default_states (np.ndarray or tf.Tensor): the default state
                attributes for a MaterialGraph
            **kwargs:
        """

        self.default_states = default_states
        super().__init__(**kwargs)

    @staticmethod
    def get_atom_features(structure) -> np.ndarray:
        """
        Get atom features from structure, may be overwritten
        Args:
            structure: (Pymatgen.Structure) pymatgen structure
        Returns:
            List of atomic numbers
        """
        if isinstance(structure, Atoms):
            return np.array(structure.get_atomic_numbers(), dtype=DataType.np_int)
        return np.array([i.specie.Z for i in structure], dtype=DataType.np_int)

    def set_default_states(self, states=None):
        """
        set the default states for MaterialGraph
        Args:
            states (np.ndarray or tf.Tensor): the state attributes
        Returns:

        """
        self.default_states = states

    def get_states(self, structure: StructureOrMolecule):
        """
        Get the state attributes from a structure
        Args:
            structure (StructureOrMolecule): structure to compute the states

        Returns: tf.Tensor or np.ndarray

        """
        states = getattr(structure, "states", None)
        if states is None:
            states = self.default_states
        return states

    @abstractmethod
    def convert(self, structure: StructureOrMolecule, **kwargs) -> MaterialGraph:
        """
        Convert the structure into a graph
        Args:
            structure:
            **kwargs:
        Returns:

        """
        pass

    def convert_many(self, structures: List[StructureOrMolecule], **kwargs) -> MaterialGraph:
        """
        Convert many structures into one single graph
        Args:
            structures: List of structures
            **kwargs:
        Returns: MaterialGraph
        """
        graphs = [self.convert(structure, **kwargs) for structure in structures]
        return assemble_material_graph(graphs)

    def __call__(self, structure: StructureOrMolecule, *args, **kwargs) -> MaterialGraph:
        """
        A thin wrapper for calling `convert` method
        Args:
            structure:
            *args:
            **kwargs:
        Returns:

        """
        return self.convert(structure)


@register
class RadiusCutoffGraphConverter(BaseGraphConverter):
    """
    Constructing a material graph with fixed radius cutoff
    """

    def __init__(
        self,
        cutoff: float = 5.0,
        has_threebody: bool = True,
        threebody_cutoff: Optional[float] = None,
        **kwargs,
    ):
        """

        Args:
            cutoff: float, cutoff radius
            atom_converter: Converter, atom converter
            bond_converter: Converter, bond converter
            three_body_cutoff: float, three-body cutoff radius
            **kwargs:
        """
        self.cutoff = cutoff
        self.has_threebody = has_threebody

        if has_threebody:
            if threebody_cutoff is None:
                threebody_cutoff = cutoff - 1.0
            if threebody_cutoff > cutoff:
                raise ValueError("Three body cutoff has to be smaller than two " "body")
        self.threebody_cutoff = threebody_cutoff

        super().__init__(**kwargs)

    def convert(self, structure: StructureOrMolecule, **kwargs) -> MaterialGraph:
        """
        Convert the structure into graph
        Args:
            structure: Structure or Molecule
        Returns
            MaterialGraph
        """

        if isinstance(structure, Atoms):
            atom_positions = np.asarray(structure.get_positions(), dtype=DataType.np_float)
        else:
            atom_positions = np.array(structure.cart_coords, dtype=DataType.np_float)
        state_attributes = self.get_states(structure)

        sender_indices, receiver_indices, images, distances = get_fixed_radius_bonding(structure, self.cutoff)

        if np.size(np.unique(sender_indices)) < len(structure):
            logger.warning("Isolated atoms found in the structure")

        bonds = distances[:, None]
        bond_atom_indices = np.array([sender_indices, receiver_indices], dtype=DataType.np_int).T
        pbc_offsets = images.astype(DataType.np_int)

        mg = MaterialGraph(  # type: ignore
            bonds=bonds,
            bond_weights=distances,
            bond_atom_indices=bond_atom_indices,  # noqa
            pbc_offsets=pbc_offsets,
        )  # noqa

        n_atom = len(structure)

        atoms = reshape_array(self.get_atom_features(structure), [n_atom, None])
        if state_attributes is not None:
            state_attributes = reshape_array(state_attributes, [1, None])

        n_bonds = np.array([mg.bonds.shape[0]], dtype=DataType.np_int)  # type: ignore
        mg = mg.replace(
            states=state_attributes,
            atoms=atoms,
            atom_positions=atom_positions,
            n_atoms=np.array([n_atom], dtype=DataType.np_int),
            n_bonds=n_bonds,
        )

        if isinstance(structure, Structure):
            mg = mg.replace(lattices=np.array(structure.lattice.matrix.reshape((1, 3, 3)), dtype=DataType.np_float))

        if isinstance(structure, Atoms):
            mg = mg.replace(lattices=np.array(structure.cell[:].reshape((1, 3, 3)), dtype=DataType.np_float))

        if self.has_threebody:
            mg = include_threebody_indices(mg, self.threebody_cutoff)
        return mg

    def __str__(self):
        s = f"<RadiusCutoffGraphConverter cutoff={self.cutoff}" f" threebody_cutoff={self.threebody_cutoff} >"
        return s

    def __repr__(self):
        return str(self)

    def get_config(self) -> Dict:
        """
        Get serialized dictionary
        Returns: Dict

        """
        return {
            "cutoff": self.cutoff,
            "threebody_cutoff": self.threebody_cutoff,
            "name": self.name,
            "trainable": self.trainable,
            "dtype": self.dtype,
            "has_threebody": self.has_threebody,
        }



import unittest

import numpy as np
import tensorflow as tf
from pymatgen.core import Lattice, Structure

from m3gnet.graph import MaterialGraph, RadiusCutoffGraphConverter
from m3gnet.layers import MLP, Embedding, GatedMLP, Pipe


class _Layer(tf.keras.layers.Layer):
    def call(self, x, **kwargs):
        return x * x


class TestCore(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        gc = RadiusCutoffGraphConverter(5)
        s1 = Structure(Lattice.cubic(3.17), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        s2 = Structure(Lattice.cubic(3), ["Mo", "Fe"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        cls.g1 = gc.convert(s1)
        cls.g2 = gc.convert(s2)
        cls.x = np.random.normal(size=(10, 4))

    def test_pipe(self):
        pipe = Pipe(layers=[_Layer(), _Layer()])
        y = pipe(self.x)
        self.assertTrue(np.linalg.norm(self.x**4 - y) < 0.001)

    def test_mlp(self):
        layer = MLP(neurons=[10, 3], activations="swish")
        out = layer(self.x)
        self.assertTupleEqual(tuple(out.shape), (10, 3))
        out2 = self.x
        for x in layer.pipe.layers:
            out2 = x(out2)
        np.testing.assert_array_almost_equal(out, out2)

    def test_gated_mlp(self):
        layer = GatedMLP(neurons=[10, 3], activations="swish")
        self.assertTrue(isinstance(layer.pipe, Pipe))
        self.assertTrue(isinstance(layer.gate, Pipe))
        self.assertTrue(layer.gate.layers[-1].activation == tf.keras.activations.sigmoid)

    def test_embedding(self):
        emb = Embedding(2, 8)
        x = np.array([[0], [1], [0], [1]])
        self.assertTupleEqual(tuple(emb(x).shape), (4, 8))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import tensorflow as tf
from pymatgen.core import Lattice, Structure

from m3gnet.graph import Index, MaterialGraph, RadiusCutoffGraphConverter


class TestConverter(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s1 = Structure(Lattice.cubic(3.17), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        cls.g1 = RadiusCutoffGraphConverter(cutoff=5.0, threebody_cutoff=4.0).convert(cls.s1)

    def test_graph(self):
        self.assertTrue(isinstance(self.g1, MaterialGraph))
        glist = self.g1.as_list()
        np.testing.assert_array_almost_equal(glist[Index.ATOMS].ravel(), [42, 42])
        gstr = str(self.g1)
        self.assertTrue(gstr.startswith("<Material"))

        self.assertTrue(isinstance(self.g1.atoms, np.ndarray))
        gtf = self.g1.as_tf()
        self.assertTrue(isinstance(gtf.atoms, tf.Tensor))
        self.assertTrue(self.g1.n_atom == 2)
        self.assertTrue(self.g1.n_bond == self.g1.n_bonds[0])
        self.assertTrue(self.g1.n_struct == 1)
        self.assertTrue(self.g1.has_threebody)

        g2 = MaterialGraph.from_list(self.g1.as_list())
        self.assertTrue(self.g1 == g2)
        g3 = self.g1.copy()
        self.assertTrue(self.g1 == g3)


if __name__ == "__main__":
    unittest.main()



"""
Computing various graph based operations
"""

from typing import List, Optional, Union

import numpy as np
import tensorflow as tf

from m3gnet.config import DataType
from m3gnet.utils import get_length, get_segment_indices_from_n

from ._threebody_indices import compute_threebody as _compute_threebody
from ._types import Index, MaterialGraph


def get_pair_vector_from_graph(graph: List):
    """
    Given a graph list return pair vectors that form the bonds
    Args:
        graph (List): graph list, obtained by MaterialGraph.as_list()

    Returns: pair vector tf.Tensor

    """
    atom_positions = graph[Index.ATOM_POSITIONS]
    lattices = graph[Index.LATTICES]
    pbc_offsets = graph[Index.PBC_OFFSETS]
    bond_atom_indices = graph[Index.BOND_ATOM_INDICES]
    n_bonds = graph[Index.N_BONDS]
    if lattices is not None:
        lattices = tf.gather(lattices, get_segment_indices_from_n(n_bonds))
        offset_vec = tf.keras.backend.batch_dot(tf.cast(pbc_offsets, DataType.tf_float), lattices)
    else:
        offset_vec = tf.constant([[0.0, 0.0, 0.0]], dtype=DataType.tf_float)
    diff = (
        tf.gather(atom_positions, bond_atom_indices[:, 1])
        + offset_vec
        - tf.gather(atom_positions, bond_atom_indices[:, 0])
    )
    return tf.cast(diff, DataType.tf_float)


def tf_compute_distance_angle(graph: List):
    """
    Given a graph with pair, triplet indices, calculate the pair distance,
    triplet angles, etc.
    Args:
        graph: MaterialGraph in List format

    Returns: MaterialGraph in List format
    """
    graph = graph[:]
    pair_vectors = get_pair_vector_from_graph(graph)
    pair_rij = get_length(pair_vectors)
    vij = tf.gather(pair_vectors, graph[Index.TRIPLE_BOND_INDICES][:, 0])
    vik = tf.gather(pair_vectors, graph[Index.TRIPLE_BOND_INDICES][:, 1])
    rij = tf.gather(pair_rij, graph[Index.TRIPLE_BOND_INDICES][:, 0])
    rik = tf.gather(pair_rij, graph[Index.TRIPLE_BOND_INDICES][:, 1])
    cos_jik = tf.reduce_sum(vij * vik, axis=1) / (rij * rik)
    cos_jik = tf.clip_by_value(cos_jik, -1, 1)
    graph[Index.BOND_WEIGHTS] = pair_rij
    graph[Index.BONDS] = pair_rij
    graph[Index.TRIPLE_BOND_LENGTHS] = rik
    graph[Index.THETA] = cos_jik  # theta is in fact costheta
    graph[Index.PHI] = tf.zeros_like(cos_jik)  # dummy phi here
    return graph


def include_threebody_indices(graph: Union[MaterialGraph, List], threebody_cutoff: Optional[float] = None):
    """
    Given a graph without threebody indices, add the threebody indices
    according to a threebody cutoff radius
    Args:
        graph: MaterialGraph
        threebody_cutoff: float, threebody cutoff radius

    Returns: new graph with added threebody indices

    """
    if isinstance(graph, MaterialGraph):
        is_graph = True
        graph_list: List = graph.as_list()
    else:
        is_graph = False
        graph_list = graph

    return _list_include_threebody_indices(graph_list, threebody_cutoff=threebody_cutoff, is_graph=is_graph)


def _list_include_threebody_indices(graph: List, threebody_cutoff: Optional[float] = None, is_graph: bool = False):
    graph = graph[:]
    bond_atom_indices = graph[Index.BOND_ATOM_INDICES]
    n_bond = bond_atom_indices.shape[0]
    if n_bond > 0 and threebody_cutoff is not None:
        valid_three_body = graph[Index.BOND_WEIGHTS] <= threebody_cutoff
        ij_reverse_map = np.where(valid_three_body)[0]
        original_index = np.arange(n_bond)[valid_three_body]
        bond_atom_indices = bond_atom_indices[valid_three_body, :]
    else:
        ij_reverse_map = None
        original_index = np.arange(n_bond)

    if bond_atom_indices.shape[0] > 0:
        bond_indices, n_triple_ij, n_triple_i, n_triple_s = _compute_threebody(
            np.ascontiguousarray(bond_atom_indices, dtype="int32"),
            np.array(graph[Index.N_ATOMS], dtype="int32"),
        )

        if ij_reverse_map is not None:
            n_triple_ij_ = np.zeros(shape=(n_bond,), dtype="int32")
            n_triple_ij_[ij_reverse_map] = n_triple_ij
            n_triple_ij = n_triple_ij_
        bond_indices = original_index[bond_indices]
        bond_indices = np.array(bond_indices, dtype="int32")
    else:
        bond_indices = np.reshape(np.array([], dtype="int32"), [-1, 2])
        if n_bond == 0:
            n_triple_ij = np.array([], dtype="int32")
        else:
            n_triple_ij = np.array([0] * n_bond, dtype="int32")
        n_triple_ij = np.array([], dtype="int32")
        n_triple_i = np.array([0] * len(graph[Index.ATOMS]), dtype="int32")
        n_triple_s = np.array([0], dtype="int32")
    graph[Index.TRIPLE_BOND_INDICES] = bond_indices
    graph[Index.N_TRIPLE_IJ] = n_triple_ij
    graph[Index.N_TRIPLE_I] = n_triple_i
    graph[Index.N_TRIPLE_S] = n_triple_s
    if is_graph:
        graph = MaterialGraph.from_list(graph)
    return graph



"""
Materials Graph Network
"""
from copy import deepcopy
from typing import List, Optional

import tensorflow as tf

from m3gnet.utils import register

from ._atom import AtomNetwork
from ._base import GraphUpdate, GraphUpdateFunc
from ._bond import BondNetwork, PairRadialBasisExpansion
from ._core import Embedding
from ._state import StateNetwork


@register
class GraphNetworkLayer(GraphUpdate):
    """
    A graph network layer features bond/atom/state update in the sequence
    bond -> atom -> state. The input and output of each step are graphs/graph
    converted lists
    """

    def __init__(
        self,
        bond_network: Optional[GraphUpdate] = None,
        atom_network: Optional[GraphUpdate] = None,
        state_network: Optional[GraphUpdate] = None,
        **kwargs,
    ):
        """

        Args:
            bond_network (GraphUpdate): bond update network
            atom_network (GraphUpdate): atom update network
            state_network (GraphUpdate): state update network
            **kwargs:
        """
        self.bond_network = bond_network or GraphUpdate()
        self.atom_network = atom_network or GraphUpdate()
        self.state_network = state_network or GraphUpdate()
        super().__init__(**kwargs)

    def call(self, graph: List, **kwargs) -> List:
        """
        Args:
            graph (List): a graph in list representation
            **kwargs:
        Returns: tf.Tensor
        """
        out = self.state_network(self.atom_network(self.bond_network(graph)))
        return out

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update(
            {
                "bond_network": self.bond_network,
                "atom_network": self.atom_network,
                "state_network": self.state_network,
            }
        )
        return config


def _get_bond_featurizer(nfeat_bond, n_bond_types, bond_embedding_dim, rbf_type, kwargs):
    # bond network settings for the featurization layer
    # no need for bond network
    if nfeat_bond is not None:
        return BondNetwork()

    # no feature length, meaning that the features are either
    # integer  types or just distances
    if bond_embedding_dim is not None:
        # bond attributes are integer types
        return GraphFieldEmbedding(
            nvocal=n_bond_types,
            embedding_dim=bond_embedding_dim,
            field="bonds",
            name="bond_embedding",
        )

    if rbf_type.lower() == "gaussian":
        centers = kwargs.pop("centers", None)
        width = kwargs.pop("width", None)
        if centers is None and width is None:
            raise ValueError(
                "If the bond attributes are single float values, "
                "we expect the value to be expanded before passing "
                "to the pretrained. Therefore, `centers` and `width` for "
                "Gaussian basis expansion are needed"
            )
        # bond attributes are distances
        bond_network = PairRadialBasisExpansion(rbf_type="Gaussian", centers=centers, width=width)
        return bond_network

    if rbf_type.lower() == "sphericalbessel":
        max_l = kwargs.pop("max_l")
        max_n = kwargs.pop("max_n")
        cutoff = kwargs.pop("cutoff")
        smooth = kwargs.pop("smooth", False)
        return PairRadialBasisExpansion(
            rbf_type="SphericalBessel",
            max_l=max_l,
            max_n=max_n,
            cutoff=cutoff,
            smooth=smooth,
        )
    raise ValueError("Cannot derive bond network type")


@register
class GraphFeaturizer(GraphNetworkLayer):
    """
    Graph featurizer that does several things to convert an initial graph
    with atomic number atom attributes and bond distance bond attributes to
    a graph with proper feature dimensions
    """

    def __init__(
        self,
        nfeat_bond: Optional[int] = None,
        nfeat_atom: Optional[int] = None,
        nfeat_state: Optional[int] = None,
        n_bond_types: Optional[int] = None,
        n_atom_types: Optional[int] = None,
        n_state_types: Optional[int] = None,
        bond_embedding_dim: Optional[int] = None,
        atom_embedding_dim: Optional[int] = None,
        state_embedding_dim: Optional[int] = None,
        rbf_type="SpherialBessel",
        **kwargs,
    ):
        """

        Args:
            nfeat_bond (int): bond feature dimension
            nfeat_atom (int): atom feature dimension
            nfeat_state (int): state feature dimension
            n_bond_types (int): number of bond types, used only when the
                bond is categorical
            n_atom_types (int): number of atom types
            n_state_types (int): number of state types, used only when state is
                of categorical type
            bond_embedding_dim (int): embedding dimension for bond
            atom_embedding_dim (int): embedding dimension for atom
            state_embedding_dim (int): embedding dimension for state
            rbf_type (str): radial basis function type, choose between
                "SphericalBessel" and "Gaussian"
            **kwargs:
        """
        self.kwargs = deepcopy(kwargs)

        # bond network settings for the featurization layer
        bond_network = _get_bond_featurizer(nfeat_bond, n_bond_types, bond_embedding_dim, rbf_type, kwargs)
        # atom network for the featurization layer
        if nfeat_atom is None:
            if n_atom_types is None or atom_embedding_dim is None:
                raise ValueError("Either specify nfeat_atom or " "n_atom_types and atom_embedding_dim")
            atom_network = GraphFieldEmbedding(
                nvocal=n_atom_types + 1,
                embedding_dim=atom_embedding_dim,
                field="atoms",
                name="atom_embedding",
            )
        else:
            atom_network = AtomNetwork()

        # state network for the featurization layer
        if nfeat_state is None and state_embedding_dim is not None:
            if n_state_types is None:
                raise ValueError("Either specify nfeat_state or " "n_state_types and state_embedding_dim")
            state_network = GraphFieldEmbedding(
                nvocal=n_state_types,
                embedding_dim=state_embedding_dim,
                field="states",
                name="state_embedding",
            )
        else:
            state_network = StateNetwork()

        self.nfeat_bond = nfeat_bond
        self.nfeat_atom = nfeat_atom
        self.nfeat_state = nfeat_state
        self.n_bond_types = n_bond_types
        self.n_atom_types = n_atom_types
        self.n_state_types = n_state_types
        self.bond_embedding_dim = bond_embedding_dim
        self.atom_embedding_dim = atom_embedding_dim
        self.state_embedding_dim = state_embedding_dim
        self.rbf_type = rbf_type
        super().__init__(bond_network=bond_network, atom_network=atom_network, state_network=state_network, **kwargs)

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns:
        """
        config = tf.keras.layers.Layer.get_config(self)
        config.update(
            nfeat_bond=self.nfeat_bond,
            nfeat_atom=self.nfeat_atom,
            nfeat_state=self.nfeat_state,
            n_bond_types=self.n_bond_types,
            n_atom_types=self.n_atom_types,
            n_state_types=self.n_state_types,
            bond_embedding_dim=self.bond_embedding_dim,
            atom_embedding_dim=self.atom_embedding_dim,
            state_embedding_dim=self.state_embedding_dim,
            rbf_type=self.rbf_type,
        )
        config.update(**self.kwargs)
        return config


@register
class GraphFieldEmbedding(GraphUpdateFunc):
    """
    Embedding the categorical field of a graph to continuous space
    """

    def __init__(self, nvocal: int = 95, embedding_dim: int = 16, field: str = "atoms", **kwargs):
        """

        Args:
            nvocal (int): number of vocabulary
            embedding_dim (int): the embedding dimension
            field (str): graph field for embedding
            **kwargs:
        """
        self.nvocal = nvocal
        self.embedding_dim = embedding_dim
        self.field = field
        self.kwargs = kwargs
        self.embedding = Embedding(nvocal, embedding_dim, **kwargs)
        super().__init__(update_func=self.embedding, update_field=field, **kwargs)

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns:
        """
        return {
            "nvocal": self.nvocal,
            "embedding_dim": self.embedding_dim,
            "field": self.field,
            "name": self.name,
            "trainable": self.trainable,
            "dtype": self.dtype,
        }



# -*- coding: utf-8 -*-
import json
import os
import unittest
from shutil import rmtree

import numpy as np
import tensorflow as tf
from monty.tempfile import ScratchDir
from pymatgen.core import Structure

from m3gnet.models import M3GNet, Potential
from m3gnet.trainers import PotentialTrainer, Trainer

DIR = os.path.dirname(os.path.abspath(__file__))


class TestTrainer(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        with open(os.path.join(DIR, "Li-O.json")) as f:
            data = json.load(f)

        cls.structures = []
        cls.energies = []
        cls.bandgaps = []
        for i, j in data.items():
            cls.structures.append(Structure.from_dict(j["structure"]))
            cls.energies.append(j["energy"])
            cls.bandgaps.append(j["band_gap"])

    def test_train_bandgap(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=True)
        trainer = Trainer(model=m3gnet, optimizer=tf.keras.optimizers.Adam(1e-2))

        trainer.train(self.structures[:30], self.bandgaps[:30], epochs=2, train_metrics=["mae"])
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))

    def test_train_energy(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=False)

        trainer = Trainer(model=m3gnet, optimizer=tf.keras.optimizers.Adam(1e-3))

        with ScratchDir("."):
            trainer.train(
                self.structures[:30],
                self.energies[:30],
                epochs=2,
                train_metrics=["mae"],
            )
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))

    def test_train_potential(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=False)
        potential = Potential(model=m3gnet)

        trainer = PotentialTrainer(potential=potential, optimizer=tf.keras.optimizers.Adam(1e-3))

        n_atoms = [len(i) for i in self.structures]
        fake_forces = [np.random.normal(size=(i, 3)) for i in n_atoms]
        fake_stress = [np.random.normal(size=(3, 3)) for i in n_atoms]
        with ScratchDir("."):
            trainer.train(
                self.structures[:30],
                self.energies[:30],
                fake_forces[:30],
                fake_stress[:30],
                validation_graphs_or_structures=self.structures[30:40],
                val_energies=self.energies[30:40],
                val_forces=fake_forces[30:40],
                val_stresses=fake_stress[30:40],
                epochs=2,
                fit_per_element_offset=True,
                save_checkpoint=False,
            )
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))

    def test_train_energy_offset(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=False)

        trainer = Trainer(model=m3gnet, optimizer=tf.keras.optimizers.Adam(1e-3))

        trainer.train(
            self.structures[:30],
            self.energies[:30],
            epochs=2,
            train_metrics=["mae"],
            fit_per_element_offset=True,
        )
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))


if __name__ == "__main__":
    unittest.main()



"""
Graph pretrained
"""
from ._base import BasePotential, GraphModelMixin, Potential
from ._dynamics import M3GNetCalculator, MolecularDynamics, Relaxer
from ._m3gnet import M3GNet

__all__ = [
    "GraphModelMixin",
    "BasePotential",
    "Potential",
    "M3GNet",
    "M3GNetCalculator",
    "Relaxer",
    "MolecularDynamics",
]



import unittest

import numpy as np
from pymatgen.core import Lattice, Structure

from m3gnet.graph import (
    MaterialGraph,
    MaterialGraphBatch,
    MaterialGraphBatchEnergyForceStress,
    RadiusCutoffGraphConverter,
    assemble_material_graph,
)


class TestBatch(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        gc = RadiusCutoffGraphConverter(5)
        s1 = Structure(Lattice.cubic(3.17), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        s2 = Structure(Lattice.cubic(3), ["Mo", "Fe"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        cls.g1 = gc.convert(s1)
        cls.g2 = gc.convert(s2)

    def test_assemble_material_graph(self):
        g_assembled = assemble_material_graph([self.g1, self.g2])
        n_bond1 = self.g1.bonds.shape[0]
        n_bond2 = self.g2.bonds.shape[0]
        n_atom1 = self.g1.atoms.shape[0]

        self.assertTrue(g_assembled.atoms.shape == (4, 1))
        self.assertTrue(g_assembled.bonds.shape[0] == n_bond1 + n_bond2)

        self.assertTrue(np.allclose(g_assembled.bond_atom_indices[:n_bond1], self.g1.bond_atom_indices))
        self.assertTrue(
            np.allclose(
                g_assembled.bond_atom_indices[n_bond1:],
                self.g2.bond_atom_indices + n_atom1,
            )
        )

        g_assembled2 = assemble_material_graph([self.g1.as_list(), self.g2.as_list()])

        g_assembled_mg = MaterialGraph.from_list(g_assembled2)
        self.assertTrue(g_assembled_mg.atoms.shape == (4, 1))
        self.assertTrue(g_assembled_mg.bonds.shape[0] == n_bond1 + n_bond2)
        self.assertTrue(np.allclose(g_assembled_mg.bond_atom_indices[:n_bond1], self.g1.bond_atom_indices))
        self.assertTrue(
            np.allclose(
                g_assembled_mg.bond_atom_indices[n_bond1:],
                self.g2.bond_atom_indices + n_atom1,
            )
        )

    def test_sequential_assemble(self):
        # test stepwise combine
        g3 = assemble_material_graph([self.g1, self.g2])
        g3 = assemble_material_graph([self.g1, g3])

        g4 = assemble_material_graph([self.g1, self.g1, self.g2])
        self.assertTrue(g3 == g4)

    def test_sequential_assemble_list(self):
        g1 = self.g1.as_list()
        g2 = self.g2.as_list()

        g3 = assemble_material_graph([g1, g2])
        g3 = assemble_material_graph([g1, g3])

        g4 = assemble_material_graph([g1, g1, g2])
        self.assertTrue(MaterialGraph.from_list(g3) == MaterialGraph.from_list(g4))

    def test_materialgraph_batch(self):
        graphs = [self.g1, self.g2]
        targets = [0.1, 0.2]
        gb = MaterialGraphBatch(graphs, targets, batch_size=2)
        first_batch = gb[0]
        self.assertTrue(isinstance(first_batch[0], MaterialGraph))
        self.assertTrue(first_batch[1].shape == (2,))
        index = gb.graph_index[:]
        gb.on_epoch_end()
        index2 = gb.graph_index[:]
        self.assertTrue(set(index) == set(index2))
        self.assertTrue(len(gb) == 1)

    def test_materialgraph_efs(self):
        graphs = [self.g1, self.g2]
        energies = [0.1, 0.2]
        forces = [np.random.normal(size=(2, 3)), np.random.normal(size=(2, 3))]
        stresses = [np.random.normal(size=(3, 3)), np.random.normal(size=(3, 3))]
        mgb = MaterialGraphBatchEnergyForceStress(
            graphs=graphs,
            energies=energies,
            forces=forces,
            stresses=stresses,
            batch_size=2,
        )
        first_batch = mgb[0]
        outshapes = [i.shape for i in first_batch[1]]
        self.assertListEqual(
            list(outshapes[0]),
            [
                2,
            ],
        )
        self.assertListEqual(list(outshapes[1]), [4, 3])
        self.assertListEqual(list(outshapes[2]), [2, 3, 3])


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

import numpy as np
from pymatgen.core import Lattice, Molecule, Structure

from m3gnet.graph import Index, RadiusCutoffGraphConverter, tf_compute_distance_angle
from m3gnet.layers import ConcatAtoms, PairRadialBasisExpansion, ThreeDInteraction


class TestAgg(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])
        mol = Molecule(["C", "O"], [[0, 0, 0], [1.1, 0, 0]])
        rc = RadiusCutoffGraphConverter(cutoff=4.0, has_threebody=True, threebody_cutoff=4.0)
        cls.struct_graph = rc.convert(s).as_list()
        cls.mol_graph = rc.convert(mol).as_list()

    def test_bond_network(self):
        prb = PairRadialBasisExpansion(rbf_type="Gaussian", centers=np.linspace(0, 4, 10), width=0.5)
        graph = prb(self.struct_graph)
        self.assertTupleEqual(tuple(graph[Index.BONDS].shape), (28, 10))
        prb = PairRadialBasisExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=False)
        graph = prb(self.struct_graph)
        self.assertTupleEqual(tuple(graph[Index.BONDS].shape), (28, 9))


if __name__ == "__main__":
    unittest.main()



"""
General utility
"""
from typing import Optional, Sequence

import numpy as np


def check_array_equal(array1: Optional[np.ndarray], array2: Optional[np.ndarray], rtol: float = 1e-3) -> bool:
    """
    Check the equality of two arrays
    Args:
        array1 (np.ndarray): first array
        array2 (np.narray): second array
        rtol (float): relative tolerance
    Returns: Bool
    """
    if array1 is None and array2 is None:
        return True
    if (array1 is None) ^ (array2 is None):
        return False
    return np.allclose(array1, array2, rtol=rtol)


def check_shape_consistency(array: Optional[np.ndarray], shape: Sequence) -> bool:
    """
    Check if array complies with shape. Shape is a sequence of
    integer that may end with None. If None is at the end of shape,
    then any shapes in array after that dimension will match with shape.

    Example: array with shape [10, 20, 30, 40] matches with [10, 20, None], but
        does not match with shape [10, 20, 30, 20]

    Args:
        array (np.ndarray or None): array to be checked
        shape (Sequence): integer array shape, it may ends with None
    Returns: bool
    """
    if array is None:
        return True
    if all(i is None for i in shape):
        return True

    array_shape = array.shape
    valid_dims = [i for i in shape if i is not None]
    n_for_check = len(valid_dims)
    return all(i == j for i, j in zip(array_shape[:n_for_check], valid_dims))


def reshape_array(array: np.ndarray, shape: Sequence) -> np.ndarray:
    """
    Take an array and reshape it according to shape. Here shape may contain
    None field at the end.

    if array shape is [3, 4] and shape is [3, 4, None], then array is shaped
    to [3, 4, 1]. If the two shapes do not match then report an error

    Args:
        array (np.ndarray): array to be reshaped
        shape (Sequence): shape dimensions

    Returns: np.ndarray, reshaped array
    """
    if not check_shape_consistency(array, shape):
        raise ValueError("array cannot be reshaped due to mismatch")

    if array.ndim >= len(shape):
        return array
    shape_r = [i if i is not None else 1 for i in shape]
    missing_dim = range(len(array.shape), len(shape_r))
    array_r = np.expand_dims(array, axis=list(missing_dim))
    tiles = [i // j for i, j in zip(shape_r, array_r.shape)]
    return np.tile(array_r, tiles)



# -*- coding: utf-8 -*-
"""
Core layers provide basic operations, e.g., MLP
"""
from typing import Dict, List, Union

import tensorflow as tf

from m3gnet.utils import register

METHOD_MAPPING = {x: eval(f"tf.math.unsorted_segment_{x}") for x in ["sum", "prod", "max", "min", "mean"]}


@register
class Pipe(tf.keras.layers.Layer):
    """
    Simple layer for consecutive layer calls, similar to Sequential
    """

    def __init__(self, layers: List, **kwargs):
        """
        Args:
            layers（list): List of layers for consecutive calls
            **kwargs:
        """
        super().__init__(**kwargs)
        self.layers = layers

    def call(self, inputs: List[tf.Tensor], **kwargs) -> tf.Tensor:
        """
        Run the inputs through the layers
        Args:
            inputs (List): a graph in list representation
            **kwargs:
        Returns: tf.Tensor
        """
        out = inputs
        for layer in self.layers:
            out = layer(out)
        return out

    def get_config(self) -> Dict:
        """
        Get layer configuration in dictionary format
        Returns: dict

        """
        config = super().get_config()
        config["layers"] = []
        for i in self.layers:
            config["layers"].append(tf.keras.layers.serialize(i))
        return config

    @classmethod
    def from_config(cls, config: Dict) -> "Pipe":
        """
        Construct Pipe object from a config dict
        Args:
            config (dict): configuration dictionary

        Returns: Pipe object

        """
        layers_dict = config.pop("layers")
        layers = [tf.keras.layers.deserialize(i) for i in layers_dict]
        return cls(layers=layers, **config)


@register
class Embedding(tf.keras.layers.Embedding):
    """
    Thin wrapper for embedding atomic numbers into feature vectors
    """

    def call(self, inputs: List[tf.Tensor]) -> tf.Tensor:
        """
        Implementation of the layer call

        Args:
            inputs (list): list representation of graph
        Returns: tf.Tensor

        """
        return super().call(inputs[:, 0])


@register
class MLP(tf.keras.layers.Layer):
    """
    Multi-layer perceptron
    """

    def __init__(
        self,
        neurons: List[int],
        activations: Union[List, str, None] = "swish",
        kernel_regularizers: Union[List, str, None] = None,
        use_bias: bool = True,
        is_output: bool = False,
        **kwargs,
    ):
        """
        Multi-layer perceptron implementation
        Args:
            neurons (list): number of neurons in each layer
            activations (list): activation functions in each layer
            kernel_regularizers (list): regularizers, list of string or the
                ones in tf.keras.regularizers
            use_bias (bool): whether to use bias in the ANN
            is_output (bool): whether the result is the final output. If it
                is the final output, we need to make sure that it gives
                float32. This is for half-precision training
            **kwargs:
        """
        super().__init__(**kwargs)
        if isinstance(activations, str) or activations is None:
            activation_list = [activations] * len(neurons)
        else:
            activation_list = activations  # type: ignore
        if isinstance(kernel_regularizers, str) or kernel_regularizers is None:
            kernel_regularizer_list = [kernel_regularizers] * len(neurons)
        else:
            kernel_regularizer_list = kernel_regularizers  # type: ignore

        self.neurons = neurons
        self.activations: Union[List, str, None] = activations
        self.kernel_regularizers: Union[List, str, None] = kernel_regularizers

        dense_layers: List[tf.keras.layers.Layer] = [
            tf.keras.layers.Dense(i, activation=j, kernel_regularizer=reg, use_bias=use_bias)
            for i, j, reg in zip(neurons, activation_list, kernel_regularizer_list)
        ]
        if is_output:
            dense_layers.append(tf.keras.layers.Activation(None, dtype="float32", name="predictions"))

        self.pipe = Pipe(layers=dense_layers)
        self.is_output = is_output
        self.use_bias = use_bias

    def call(self, inputs: List[tf.Tensor], **kwargs) -> tf.Tensor:
        """
        Implementation of the layer call

        Args:
            inputs (list): list representation of graph
        Returns: tf.Tensor

        """
        return self.pipe.call(inputs, **kwargs)

    def get_config(self) -> Dict:
        """
        Get layer configuration in dictionary format
        Returns: dict

        """
        config = super().get_config()
        config.update(
            {
                "neurons": self.neurons,
                "activations": self.activations,
                "kernel_regularizers": self.kernel_regularizers,
                "is_output": self.is_output,
                "use_bias": self.use_bias,
            }
        )
        return config


@register
class GatedMLP(tf.keras.layers.Layer):
    r"""
    Gated MLP implementation. It implements the following
        `out = MLP(x) * MLP_\\sigmoid(x)`
    where that latter changes the last layer activation function into sigmoid.
    """

    def __init__(
        self,
        neurons: List[int],
        activations: Union[List, str, None] = "swish",
        kernel_regularizers: Union[List, str, None] = None,
        use_bias: bool = True,
        **kwargs,
    ):
        """

        Args:
            neurons (list): numbrer of neurons in each layer
            activations (list): activation function in each layer
            kernel_regularizers (list): regularizer in each layer
            use_bias (bool): whether to use bias
            **kwargs:
        """
        super().__init__(**kwargs)
        if isinstance(activations, str) or activations is None:
            activation_list = [activations] * len(neurons)
        else:
            activation_list = activations[:]

        if isinstance(kernel_regularizers, str) or kernel_regularizers is None:
            kernel_regularizer_list = [kernel_regularizers] * len(neurons)
        else:
            kernel_regularizer_list = kernel_regularizers
        self.neurons = neurons
        self.activations = activations
        self.kernel_regularizers = kernel_regularizers
        dense_layers = [
            tf.keras.layers.Dense(i, activation=j, kernel_regularizer=reg, use_bias=use_bias)
            for i, j, reg in zip(neurons, activation_list, kernel_regularizer_list)
        ]
        self.pipe = Pipe(layers=dense_layers)
        activation_list[-1] = "sigmoid"
        gate_layers = [
            tf.keras.layers.Dense(i, activation=j, kernel_regularizer=reg, use_bias=use_bias)
            for i, j, reg in zip(neurons, activation_list, kernel_regularizer_list)
        ]
        self.gate = Pipe(gate_layers)
        self.use_bias = use_bias

    def call(self, inputs: List[tf.Tensor], **kwargs) -> tf.Tensor:
        """
        Implementation of the layer call

        Args:
            inputs (list): list representation of graph
        Returns: tf.Tensor

        """
        return self.pipe.call(inputs, **kwargs) * self.gate.call(inputs, **kwargs)

    def get_config(self) -> Dict:
        """
        Get layer configuration in dictionary format
        Returns: dict

        """
        config = super().get_config()
        config.update(
            {
                "neurons": self.neurons,
                "activations": self.activations,
                "kernel_regularizers": self.kernel_regularizers,
                "use_bias": self.use_bias,
            }
        )
        return config



"""
Edge networks.

The edge networks take a graph as inputs and calculate the updated
edge attributes
"""

from typing import List

import tensorflow as tf

from m3gnet.graph import Index
from m3gnet.utils import broadcast_states_to_bonds, get_segment_indices_from_n, register

from ._base import GraphUpdate
from ._core import GatedMLP
from ._basis import RBF_ALLOWED, RadialBasisFunctions


@register
class BondNetwork(GraphUpdate):
    """
    Edge network that takes a graph as input and then calculate new
    bond attributes
    """

    def update_bonds(self, graph: List) -> tf.Tensor:
        """
        Update bond info in the graph
        Args:
            graph (list): list representation of a graph
        Returns: tf.Tensor
        """
        return graph[Index.BONDS]

    def call(self, graph: List, **kwargs) -> List:
        """
        Update the bond and return a copy of the graph
        Args:
            graph (list): graph list representation
            **kwargs:
        Returns: graph list representation
        """
        graph = graph[:]
        bonds = self.update_bonds(graph)
        graph[Index.BONDS] = bonds
        return graph


def _unity_weights(graph):
    return tf.ones_like(graph[Index.BONDS][:, :1])


def _bonds_to_weights(graph):
    return graph[Index.BONDS]


@register
class PairRadialBasisExpansion(BondNetwork):
    """
    Expand the radial distance onto a basis
    """

    def __init__(self, rbf_type: str = "Gaussian", **kwargs):
        """

        Args:
            rbf_type (str): RBF type, choose from "Gaussian" and
            "SphericalBessel"
            **kwargs (dict): the necessary parameters for initialize the RBF
        """
        self.rbf_type = rbf_type
        rbf_kwds = {}
        keys = []
        for k in kwargs.keys():
            if k in RBF_ALLOWED[rbf_type]["params"]:  # type: ignore
                keys.append(k)
        for k in keys:
            rbf_kwds.update({k: kwargs.pop(k)})
        self.rbf_kwds = rbf_kwds
        self.rbf = RadialBasisFunctions(rbf_type=rbf_type, **rbf_kwds)
        if rbf_type == "Gaussian":
            self.weight_func = _unity_weights
        else:
            self.weight_func = _bonds_to_weights
        super().__init__(**kwargs)

    def update_bonds(self, graph: List) -> tf.Tensor:
        """
        Update the bond info using RBF
        Args:
            graph (list): list representation of a graph

        Returns: updated bond info
        """
        return self.rbf(tf.squeeze(graph[Index.BONDS]))  # noqa

    def call(self, graph: List, **kwargs):
        """
        On top of the changing bond, also add bond weight for later use
        Args:
            graph (list): list representation of a graph
            **kwargs:

        Returns:
        """
        graph = graph[:]
        bonds = self.update_bonds(graph)
        graph[Index.BONDS] = bonds
        graph[Index.BOND_WEIGHTS] = self.weight_func(graph)
        return graph

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update({"rbf_type": self.rbf_type})
        config.update(**self.rbf_kwds)
        return config


@register
class ConcatAtoms(BondNetwork):
    r"""
    .. math::
        eij^\prime = Update(vi⊕vj⊕eij⊕u)
    """

    def __init__(self, neurons: List[int], activation: str = "swish", **kwargs):
        """
        Concatenate the atom attributes and bond attributes (and optionally
        the state attributes, and then pass to an update function, e.g., an MLP

        Args:
            neurons (list): number of neurons in each layer
            activation (str): activation function
        """
        self.neurons = neurons
        self.activation = activation
        self.update_func = GatedMLP(neurons=neurons, activations=[activation] * len(neurons))
        self.weight_func = tf.keras.layers.Dense(neurons[-1], use_bias=False)
        super().__init__(**kwargs)

    def update_bonds(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Update bond information
        Args:
            graph (list): list representation of the graph
            **kwargs:
        Returns:
        """
        n_bonds = tf.shape(graph[Index.BOND_ATOM_INDICES])[0]
        atoms = tf.reshape(tf.gather(graph[Index.ATOMS], graph[Index.BOND_ATOM_INDICES]), (n_bonds, -1))
        if graph[Index.STATES] is None:
            concat = tf.concat([atoms, graph[Index.BONDS]], axis=-1)
        else:
            states = broadcast_states_to_bonds(graph)
            concat = tf.concat([atoms, graph[Index.BONDS], states], axis=-1)
        return self.update_func(concat) * self.weight_func(graph[Index.BOND_WEIGHTS]) + graph[Index.BONDS]

    def get_config(self) -> dict:
        """
        get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update({"neurons": self.neurons, "activation": self.activation})
        return config


@register
class ThreeDInteraction(tf.keras.layers.Layer):
    """
    Include 3D interactions to the bond update
    """

    def __init__(self, update_network: tf.keras.layers.Layer, update_network2: tf.keras.layers.Layer, **kwargs):
        """
        Args:
            update_network (tf.keras.layers.Layer): keras layer for update
                the atom attributes before merging with 3d interactions
            update_network2 (tf.keras.layers.Layer): keras layer for update
                the bond information after merging with 3d interactions
            **kwargs:
        """
        super().__init__(**kwargs)
        self.update_network = update_network
        self.update_network2 = update_network2

    def call(self, graph: List, three_basis: tf.Tensor, three_cutoff: float, **kwargs) -> List:
        """

        Args:
            graph (list): graph list representation
            three_basis (tf.Tensor): three body basis expansion
            three_cutoff (float): cutoff radius
            **kwargs:
        Returns:
        """
        graph = graph[:]
        end_atom_index = tf.gather(graph[Index.BOND_ATOM_INDICES][:, 1], graph[Index.TRIPLE_BOND_INDICES][:, 1])
        atoms = self.update_network(graph[Index.ATOMS])
        atoms = tf.gather(atoms, end_atom_index)
        basis = three_basis * atoms
        n_bonds = tf.reduce_sum(graph[Index.N_BONDS])
        weights = tf.reshape(tf.gather(three_cutoff, graph[Index.TRIPLE_BOND_INDICES]), (-1, 2))
        weights = tf.math.reduce_prod(weights, axis=-1)
        basis = basis * weights[:, None]
        new_bonds = tf.math.unsorted_segment_sum(
            basis,
            get_segment_indices_from_n(graph[Index.N_TRIPLE_IJ]),
            num_segments=n_bonds,
        )
        graph[Index.BONDS] = graph[Index.BONDS] + self.update_network2(new_bonds)
        return graph

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update(
            {
                "update_network": self.update_network,
                "update_network2": self.update_network2,
                "out_unit": self.out_unit,
            }
        )
        return config



"""
Structure related calculations and etc.
"""

from typing import Tuple

import numpy as np
from ase import Atoms
from pymatgen.core import Molecule, Structure
from pymatgen.optimization.neighbors import find_points_in_spheres

from m3gnet.config import DataType
from m3gnet.type import StructureOrMolecule


def get_fixed_radius_bonding(
    structure: StructureOrMolecule, cutoff: float = 5.0, numerical_tol: float = 1e-8
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Get graph representations from structure within cutoff
    Args:
        structure (pymatgen Structure or molecule)
        cutoff (float): cutoff radius
        numerical_tol (float): numerical tolerance

    Returns:
        center_indices, neighbor_indices, images, distances
    """
    if isinstance(structure, Structure):
        lattice_matrix = np.ascontiguousarray(np.array(structure.lattice.matrix), dtype=float)
        pbc = np.array([1, 1, 1], dtype=int)
        cart_coords = np.ascontiguousarray(np.array(structure.cart_coords), dtype=float)
    elif isinstance(structure, Molecule):
        lattice_matrix = np.array([[1000.0, 0.0, 0.0], [0.0, 1000.0, 0.0], [0.0, 0.0, 1000.0]], dtype=float)
        pbc = np.array([0, 0, 0], dtype=int)
        cart_coords = np.ascontiguousarray(np.array(structure.cart_coords), dtype=float)

    elif isinstance(structure, Atoms):
        pbc = np.array(structure.pbc, dtype=int)
        if np.all(pbc < 0.1):
            lattice_matrix = np.array(
                [[1000.0, 0.0, 0.0], [0.0, 1000.0, 0.0], [0.0, 0.0, 1000.0]],
                dtype=float,
            )
        else:
            lattice_matrix = np.ascontiguousarray(structure.cell[:], dtype=float)

        cart_coords = np.ascontiguousarray(np.array(structure.positions), dtype=float)

    else:
        raise ValueError("structure type not supported")
    r = float(cutoff)

    center_indices, neighbor_indices, images, distances = find_points_in_spheres(
        cart_coords,
        cart_coords,
        r=r,
        pbc=pbc,
        lattice=lattice_matrix,
        tol=numerical_tol,
    )
    center_indices = center_indices.astype(DataType.np_int)
    neighbor_indices = neighbor_indices.astype(DataType.np_int)
    images = images.astype(DataType.np_int)
    distances = distances.astype(DataType.np_float)
    exclude_self = (center_indices != neighbor_indices) | (distances > numerical_tol)
    return (
        center_indices[exclude_self],
        neighbor_indices[exclude_self],
        images[exclude_self],
        distances[exclude_self],
    )
